/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/03/30
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.dialect.internal.impl;

import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.JiemamyFactory;
import org.jiemamy.dialect.DataTypeResolver;
import org.jiemamy.dialect.ImportMetadataConfig;
import org.jiemamy.dialect.internal.EntityImportVisitor;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.adapter.PrecisionedDataTypeAdapter;
import org.jiemamy.model.datatype.adapter.SizedDataTypeAdapter;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.ViewModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.utils.ForEachUtil;
import org.jiemamy.utils.JmIOUtil;
import org.jiemamy.utils.metadata.ColumnMeta;
import org.jiemamy.utils.metadata.PrimaryKeyMeta;
import org.jiemamy.utils.metadata.TableMeta;
import org.jiemamy.utils.metadata.TypeSafeDatabaseMetaData;
import org.jiemamy.utils.metadata.ColumnMeta.Nullable;
import org.jiemamy.utils.metadata.TypeSafeDatabaseMetaData.TypeSafeResultSet;
import org.jiemamy.utils.visitor.AbstractCollectionVisitor;
import org.jiemamy.utils.visitor.AbstractTypeSafeResultSetVisitor;

/**
 * {@link EntityImportVisitor}のデフォルト実装クラス。
 * 
 * @author daisuke
 */
public class DefaultEntityImportVisitor extends AbstractCollectionVisitor<TableMeta, Void, SQLException> implements
		EntityImportVisitor {
	
	private static Logger logger = LoggerFactory.getLogger(DefaultEntityImportVisitor.class);
	
	private static final String[] DEFAULT_TYPES = new String[] {
		"TABLE",
		"VIEW"
	};
	
	/** DBメタデータ */
	private TypeSafeDatabaseMetaData meta;
	
	/** 書き込み先モデル */
	private RootModel rootModel;
	
	/**
	 * 読み込み対象のエンティティ型リスト
	 * 
	 * 例えば "TABLE" とか "VIEW" とか。
	 */
	private List<String> entityTypes;
	
	/**
	 * 読み込み対象のエンティティ名のリスト
	 * 
	 * ただし、空リストだった場合、全てのエンティティを読むことを意味する。
	 */
	private Collection<String> selectedEntities;
	
	private Map<String, EntityModel> importedEntities;
	
	/** データ型リゾルバ */
	private DataTypeResolver dataTypeResolver;
	
	private ImportMetadataConfig config;
	

	public void intialize(TypeSafeDatabaseMetaData meta, RootModel rootModel, ImportMetadataConfig config,
			Map<String, EntityModel> importedEntities, DataTypeResolver dataTypeResolver) {
		Validate.notNull(meta);
		Validate.notNull(rootModel);
		Validate.notNull(config);
		Validate.notNull(importedEntities);
		Validate.notNull(dataTypeResolver);
		
		this.meta = meta;
		this.rootModel = rootModel;
		this.config = config;
		this.importedEntities = importedEntities;
		this.dataTypeResolver = dataTypeResolver;
		String[] entityTypesValue = config.getEntityTypes();
		if (entityTypesValue == null) {
			entityTypesValue = DEFAULT_TYPES;
		}
		entityTypes = Arrays.asList(entityTypesValue);
		String[] selectedEntitiesValue = config.getSelectedEntities();
		if (ArrayUtils.isEmpty(selectedEntitiesValue)) {
			selectedEntities = null;
		} else {
			selectedEntities = Arrays.asList(selectedEntitiesValue);
		}
	}
	
	public Void visit(TableMeta tableMeta) throws SQLException {
		Validate.notNull(tableMeta);
		if (entityTypes.contains(tableMeta.tableType)
				&& (selectedEntities == null || selectedEntities.contains(tableMeta.tableName))) {
			JiemamyFactory factory = rootModel.getJiemamy().getFactory();
			EntityModel entityModel = createEntity(factory, tableMeta);
			if (entityModel != null) {
				rootModel.getEntities().add(entityModel);
				importedEntities.put(entityModel.getName(), entityModel);
			}
		}
		return null;
	}
	
	/**
	 * {@link DatabaseMetaData}から得られた結果を読み込み、{@link ViewModel}を生成する。
	 * 
	 * @param factory 新規モデル作成を行うためのファクトリ
	 * @param tableMeta 読み込み対象エンティティの情報
	 * @return 読み込んだ結果生成された{@link ViewModel}
	 * @throws SQLException SQLの実行に失敗した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected EntityModel createEntity(JiemamyFactory factory, TableMeta tableMeta) throws SQLException {
		Validate.notNull(factory);
		Validate.notNull(tableMeta);
		EntityModel result = null;
		
		logger.debug("type = " + tableMeta.tableType + "(" + tableMeta.tableName + ")");
		if ("TABLE".equals(tableMeta.tableType)) {
			result = createTable(factory, tableMeta.tableName);
		} else if ("VIEW".equals(tableMeta.tableType)) {
			result = createView(factory, tableMeta.tableName);
		}
		if (StringUtils.isEmpty(tableMeta.remarks) == false) {
			result.setDescription(tableMeta.remarks);
		}
		return result;
	}
	
	/**
	 * {@link DatabaseMetaData}から得られる情報を読み込み、{@link TableModel}を生成する。
	 * 
	 * @param factory 新規モデル作成を行うためのファクトリ
	 * @param tableName 読み込み対象テーブル名
	 * @return 読み込んだ結果生成された{@link TableModel}
	 * @throws SQLException SQLの実行に失敗した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected TableModel createTable(final JiemamyFactory factory, String tableName) throws SQLException {
		Validate.notNull(factory);
		Validate.notNull(tableName);
		final TableModel tableModel = factory.newModel(TableModel.class);
		final List<ColumnModel> columns = new ArrayList<ColumnModel>();
		
		tableModel.setName(tableName);
		
		TypeSafeResultSet<ColumnMeta> columnsResult = null;
		TypeSafeResultSet<PrimaryKeyMeta> keysResult = null;
		try {
			columnsResult = meta.getColumns("", config.getSchema(), tableName, "%");
			ForEachUtil.accept(columnsResult, new ColumnMetaVisitor(factory, dataTypeResolver, columns));
			
			keysResult = meta.getPrimaryKeys("", config.getSchema(), tableName);
			ForEachUtil.accept(keysResult, new PrimaryKeyMetaVisitor(factory, tableModel.getAttributes(), columns));
		} finally {
			JmIOUtil.closeQuietly(columnsResult);
			JmIOUtil.closeQuietly(keysResult);
		}
		
		for (ColumnModel columnModel : columns) {
			tableModel.getAttributes().add(columnModel);
		}
		return tableModel;
	}
	
	/**
	 * {@link DatabaseMetaData}から得られる情報を読み込み、{@link ViewModel}を生成する。
	 * 
	 * @param factory 新規モデル作成を行うためのファクトリ
	 * @param viewName 読み込み対象ビュー名
	 * @return 読み込んだ結果生成された{@link ViewModel}
	 * @throws SQLException SQLの実行に失敗した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected ViewModel createView(JiemamyFactory factory, String viewName) throws SQLException {
		Validate.notNull(factory);
		Validate.notNull(viewName);
		// UNDONE ビューの定義SQLを取得する一般的な手段が分からない。
		// bacchusでは実は実装できなかった所。
		String definition = "VIEW DEFINITION (not implemented)";
		ViewModel view = factory.newModel(ViewModel.class);
		view.setName(viewName);
		view.setDefinition(definition);
		
		return view;
	}
	
	/**
	 * データ型リゾルバを取得する。
	 * 
	 * @return データ型リゾルバ
	 */
	protected DataTypeResolver getDataTypeResolver() {
		return dataTypeResolver;
	}
	

	private static class ColumnMetaVisitor extends
			AbstractTypeSafeResultSetVisitor<ColumnMeta, List<ColumnModel>, SQLException> {
		
		private final JiemamyFactory factory;
		
		private final List<ColumnModel> columns;
		
		private final DataTypeResolver dataTypeResolver;
		

		/**
		 * インスタンスを生成する。
		 * 
		 * @param factory 新規モデル作成を行うためのファクトリ
		 * @param dataTypeResolver データ型リゾルバ
		 * @param columns 読み込んだカラムを格納するリスト
		 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
		 */
		private ColumnMetaVisitor(JiemamyFactory factory, DataTypeResolver dataTypeResolver, List<ColumnModel> columns) {
			Validate.notNull(factory);
			Validate.notNull(dataTypeResolver);
			Validate.notNull(columns);
			this.factory = factory;
			this.dataTypeResolver = dataTypeResolver;
			this.columns = columns;
		}
		
		public List<ColumnModel> visit(ColumnMeta element) {
			Validate.notNull(element);
			BuiltinDataType dataType = dataTypeResolver.resolveDataType(factory, element.dataType, element.typeName);
			
			ColumnModel columnModel = factory.newModel(ColumnModel.class);
			columnModel.setName(element.columnName);
			columnModel.setDataType(dataType);
			columns.add(columnModel);
			
			if (dataType.hasAdapter(SizedDataTypeAdapter.class)) {
				dataType.getAdapter(SizedDataTypeAdapter.class).setSize(element.columnSize);
				logger.debug("  " + element.columnName + " = " + element.typeName + "(" + element.columnSize + ")");
			} else if (dataType.hasAdapter(PrecisionedDataTypeAdapter.class)) {
				dataType.getAdapter(PrecisionedDataTypeAdapter.class).setPrecision(element.columnSize);
				dataType.getAdapter(PrecisionedDataTypeAdapter.class).setScale(element.decimalDigits);
				logger.debug("  " + element.columnName + " = " + element.typeName + "(" + element.columnSize + ", "
						+ element.decimalDigits + ")");
			} else {
				logger.debug("  " + element.columnName + " = " + element.typeName);
			}
			
			columnModel.setDefaultValue(element.columnDef);
			
			if (element.nullable == Nullable.NO_NULLS) {
				NotNullConstraint notNullConstraint = factory.newModel(NotNullConstraint.class);
				columnModel.setNotNullConstraint(notNullConstraint);
			}
			
			// TODO check制約のインポート
			
			finalResult.add(columnModel);
			return null;
		}
		
		@Override
		protected void init() {
			finalResult = CollectionsUtil.newArrayList();
		}
	}
	
	private static class PrimaryKeyMetaVisitor extends
			AbstractTypeSafeResultSetVisitor<PrimaryKeyMeta, Boolean, SQLException> {
		
		private final List<AttributeModel> attributes;
		
		private final JiemamyFactory factory;
		
		private final List<ColumnModel> columns;
		
		private PrimaryKey primaryKey;
		

		/**
		 * インスタンスを生成する。
		 * 
		 * @param factory 新規モデル作成を行うためのファクトリ
		 * @param attributes 読み込んだPKを格納する属性のリスト
		 * @param columns カラムのリスト
		 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
		 */
		private PrimaryKeyMetaVisitor(JiemamyFactory factory, List<AttributeModel> attributes, List<ColumnModel> columns) {
			Validate.notNull(factory);
			Validate.notNull(attributes);
			Validate.notNull(columns);
			this.attributes = attributes;
			this.factory = factory;
			this.columns = columns;
		}
		
		public Boolean visit(PrimaryKeyMeta element) {
			Validate.notNull(element);
			
			if (primaryKey == null) {
				primaryKey = factory.newModel(PrimaryKey.class);
				attributes.add(primaryKey);
			}
			
			for (ColumnModel columnModel : columns) {
				if (columnModel.getName().equals(element.columnName)) {
					ColumnRef columnRef = factory.newReference(columnModel);
					primaryKey.getKeyColumns().add(columnRef);
					break;
				}
			}
			return null;
		}
	}
	
}
