/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/02/10
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.dialect.internal.impl;

import java.util.List;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.ReferenceResolver;
import org.jiemamy.dialect.DataTypeResolver;
import org.jiemamy.dialect.EmitConfig;
import org.jiemamy.dialect.internal.SqlEmitter;
import org.jiemamy.dialect.internal.TokenResolver;
import org.jiemamy.internal.model.sql.Identifier;
import org.jiemamy.internal.model.sql.Keyword;
import org.jiemamy.internal.model.sql.Literal;
import org.jiemamy.internal.model.sql.Separator;
import org.jiemamy.internal.model.sql.SqlStatementImpl;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.CheckConstraint;
import org.jiemamy.model.attribute.constraint.ColumnCheckConstraint;
import org.jiemamy.model.attribute.constraint.ConstraintModel;
import org.jiemamy.model.attribute.constraint.Deferrability;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.TableCheckConstraint;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.LiteralType;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.model.entity.ViewModel;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.model.index.IndexColumnModel.SortOrder;
import org.jiemamy.model.sql.SqlStatement;
import org.jiemamy.model.sql.Token;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.utils.Disablable;
import org.jiemamy.utils.EntityDependencyCalculator;

/**
 * {@link SqlEmitter}の標準実装クラス。
 * 
 * @author daisuke
 */
public class DefaultSqlEmitter implements SqlEmitter {
	
	private final DataTypeResolver dataTypeResolver;
	
	private TokenResolver tokenResolver = new DefaultTokenResolver();
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param dataTypeResolver データ型リゾルバ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public DefaultSqlEmitter(DataTypeResolver dataTypeResolver) {
		Validate.notNull(dataTypeResolver);
		this.dataTypeResolver = dataTypeResolver;
	}
	
	public List<SqlStatement> emit(RootModel rootModel, EmitConfig config) {
		Validate.notNull(rootModel);
		Validate.notNull(config);
		List<SqlStatement> result = CollectionsUtil.newArrayList();
		
		// THINK RootBeginScript と CreateSchemaStatement、どっちが先？
		
		if (StringUtils.isEmpty(rootModel.getBeginScript()) == false) {
			result.add(new SqlStatementImpl(Literal.of(rootModel.getBeginScript(), LiteralType.FRAGMENT)));
		}
		
		if (StringUtils.isEmpty(rootModel.getSchemaName()) == false && config.emitCreateSchemaStatement()) {
			if (config.emitDropStatements()) {
				result.add(emitDropSchemaStatement(rootModel.getSchemaName()));
			}
			
			result.add(emitCreateSchemaStatement(rootModel.getSchemaName()));
		}
		
		if (config.emitDropStatements()) {
			List<EntityModel> dropList = EntityDependencyCalculator.getSortedEntityList(rootModel);
			EntityModel[] array = dropList.toArray(new EntityModel[dropList.size()]);
			ArrayUtils.reverse(array);
			for (EntityModel entityModel : array) {
				result.add(emitDropEntityStatement(entityModel));
			}
		}
		
		for (EntityModel entityModel : EntityDependencyCalculator.getSortedEntityList(rootModel)) {
			if (entityModel.hasAdapter(Disablable.class)
					&& Boolean.TRUE.equals(entityModel.getAdapter(Disablable.class).isDisabled())) {
				continue;
			}
			if (StringUtils.isEmpty(entityModel.getBeginScript()) == false) {
				result.add(new SqlStatementImpl(Literal.of(entityModel.getBeginScript(), LiteralType.FRAGMENT)));
			}
			
			result.add(emitCreateStatement(entityModel));
			
			if (entityModel instanceof TableModel) {
				TableModel tableModel = (TableModel) entityModel;
				for (IndexModel indexModel : tableModel.getIndexes()) {
					if (indexModel.hasAdapter(Disablable.class)
							&& Boolean.TRUE.equals(indexModel.getAdapter(Disablable.class).isDisabled())) {
						continue;
					}
					if (config.emitDropStatements()) {
						result.add(emitDropIndexStatement(indexModel));
					}
					
					result.add(emitCreateIndexStatement(tableModel, indexModel));
				}
			}
			
			if (StringUtils.isEmpty(entityModel.getEndScript()) == false) {
				result.add(new SqlStatementImpl(Literal.of(entityModel.getEndScript(), LiteralType.FRAGMENT)));
			}
		}
		
		int dataSetIndex = config.getDataSetIndex();
		if (dataSetIndex >= 0 && dataSetIndex < rootModel.getDataSets().size()) {
			Jiemamy jiemamy = rootModel.getJiemamy();
			JiemamyFactory factory = jiemamy.getFactory();
			
			DataSetModel dataSetModel = rootModel.getDataSets().get(dataSetIndex);
			
			result.add(new SqlStatementImpl(Keyword.of("BEGIN"), Separator.SEMICOLON));
			for (EntityModel entityModel : EntityDependencyCalculator.getSortedEntityList(rootModel)) {
				if (entityModel.hasAdapter(Disablable.class)
						&& Boolean.TRUE.equals(entityModel.getAdapter(Disablable.class).isDisabled())) {
					continue;
				}
				if (entityModel instanceof TableModel) {
					TableModel tableModel = (TableModel) entityModel;
					TableRef tableRef = factory.newReference(tableModel);
					List<RecordModel> records = dataSetModel.getRecords().get(tableRef);
					if (records != null) {
						for (RecordModel recordModel : records) {
							result.add(emitInsertStatement(tableModel, recordModel));
						}
					}
				}
				
			}
			result.add(new SqlStatementImpl(Keyword.of("COMMIT"), Separator.SEMICOLON));
		}
		
		if (StringUtils.isEmpty(rootModel.getEndScript()) == false) {
			result.add(new SqlStatementImpl(Literal.of(rootModel.getEndScript(), LiteralType.FRAGMENT)));
		}
		
		return result;
	}
	
	/**
	 * CREATE INDEX文を出力する。
	 * 
	 * @param tableModel インデックスをつけるテーブル
	 * @param indexModel 対象インデックス
	 * @return CREATE INDEX文
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected SqlStatement emitCreateIndexStatement(TableModel tableModel, IndexModel indexModel) {
		Validate.notNull(tableModel);
		Validate.notNull(indexModel);
		List<Token> tokens = CollectionsUtil.newArrayList();
		tokens.add(Keyword.CREATE);
		if (indexModel.isUnique()) {
			tokens.add(Keyword.UNIQUE);
		}
		tokens.add(Keyword.INDEX);
		tokens.add(Identifier.of(indexModel.getName()));
		tokens.add(Keyword.ON);
		tokens.add(Identifier.of(tableModel.getName()));
		tokens.add(Separator.LEFT_PAREN);
		
		List<IndexColumnModel> indexColumns = indexModel.getIndexColumns();
		for (IndexColumnModel indexColumnModel : indexColumns) {
			tokens.addAll(emitIndexColumnClause(indexColumnModel));
			tokens.add(Separator.COMMA);
		}
		
		if (indexModel.getIndexColumns().isEmpty() == false) {
			tokens.remove(tokens.size() - 1);
		}
		tokens.add(Separator.RIGHT_PAREN);
		tokens.add(Separator.SEMICOLON);
		return new SqlStatementImpl(tokens);
	}
	
	/**
	 * CREATE SCHEMA文を出力する。
	 * 
	 * @param schemaName スキーマ名
	 * @return CREATE SCHEMA文
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected SqlStatement emitCreateSchemaStatement(String schemaName) {
		Validate.notNull(schemaName);
		List<Token> tokens = CollectionsUtil.newArrayList();
		tokens.add(Keyword.CREATE);
		tokens.add(Keyword.SCHEMA);
		tokens.add(Identifier.of(schemaName));
		tokens.add(Separator.SEMICOLON);
		return new SqlStatementImpl(tokens);
	}
	
	/**
	 * DDLを出力する。
	 * 
	 * @param entityModel 対象エンティティ
	 * @return DDL
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected SqlStatement emitCreateStatement(EntityModel entityModel) {
		Validate.notNull(entityModel);
		return EntityEmitStrategy.fromEntity(entityModel).emit(entityModel, dataTypeResolver, tokenResolver);
	}
	
	/**
	 * エンティティのDROP文を出力する。
	 * 
	 * @param entityModel 対象エンティティ
	 * @return DROP文
	 */
	protected SqlStatement emitDropEntityStatement(EntityModel entityModel) {
		List<Token> tokens = CollectionsUtil.newArrayList();
		tokens.add(Keyword.DROP);
		tokens.addAll(tokenResolver.resolve(entityModel));
		tokens.add(Identifier.of(entityModel.getName()));
		tokens.add(Separator.SEMICOLON);
		return new SqlStatementImpl(tokens);
	}
	
	/**
	 * インデックスのDROP文を出力する。
	 * 
	 * @param indexModel 対象インデックス
	 * @return DROP文
	 */
	protected SqlStatement emitDropIndexStatement(IndexModel indexModel) {
		List<Token> tokens = CollectionsUtil.newArrayList();
		tokens.add(Keyword.DROP);
		tokens.add(Keyword.INDEX);
		tokens.add(Identifier.of(indexModel.getName()));
		tokens.add(Separator.SEMICOLON);
		return new SqlStatementImpl(tokens);
	}
	
	/**
	 * スキーマのDROP文を出力する。
	 * 
	 * @param schemaName スキーマ名
	 * @return DROP文
	 */
	protected SqlStatement emitDropSchemaStatement(String schemaName) {
		List<Token> tokens = CollectionsUtil.newArrayList();
		tokens.add(Keyword.DROP);
		tokens.add(Keyword.SCHEMA);
		tokens.add(Identifier.of(schemaName));
		tokens.add(Separator.SEMICOLON);
		return new SqlStatementImpl(tokens);
	}
	
	/**
	 * インデックスカラムの定義句を出力する。
	 * 
	 * @param indexColumnModel 対象インデックスカラム
	 * @return トークンシーケンス
	 */
	protected List<Token> emitIndexColumnClause(IndexColumnModel indexColumnModel) {
		List<Token> tokens = CollectionsUtil.newArrayList();
		ReferenceResolver referenceResolver = indexColumnModel.getJiemamy().getReferenceResolver();
		ColumnRef columnRef = indexColumnModel.getColumnRef();
		ColumnModel columnModel = referenceResolver.resolve(columnRef);
		tokens.add(Identifier.of(columnModel.getName()));
		SortOrder sortOrder = indexColumnModel.getSortOrder();
		tokens.addAll(tokenResolver.resolve(sortOrder));
		return tokens;
	}
	
	/**
	 * INSERT文を出力する。
	 * 
	 * @param tableModel 対象テーブル
	 * @param recordModel 対象レコード
	 * @return INSERT文
	 */
	protected SqlStatement emitInsertStatement(TableModel tableModel, RecordModel recordModel) {
		JiemamyFactory factory = tableModel.getJiemamy().getFactory();
		List<Token> tokens = CollectionsUtil.newArrayList();
		tokens.add(Keyword.INSERT);
		tokens.add(Keyword.INTO);
		tokens.add(Identifier.of(tableModel.getName()));
		
		List<Token> columnList = CollectionsUtil.newArrayList();
		List<Token> dataList = CollectionsUtil.newArrayList();
		columnList.add(Separator.LEFT_PAREN);
		dataList.add(Separator.LEFT_PAREN);
		
		List<ColumnModel> columns = tableModel.findColumns();
		for (ColumnModel columnModel : columns) {
			columnList.add(Identifier.of(columnModel.getName()));
			columnList.add(Separator.COMMA);
			
			ReferenceResolver referenceResolver = columnModel.getJiemamy().getReferenceResolver();
			BuiltinDataType builtinDataType = columnModel.getDataType().toBuiltinDataType(referenceResolver);
			String string = recordModel.getValues().get(factory.newReference(columnModel));
			dataList.add(Literal.of(string, builtinDataType.getCategory().getLiteralType()));
			dataList.add(Separator.COMMA);
		}
		columnList.remove(columnList.size() - 1);
		dataList.remove(dataList.size() - 1);
		columnList.add(Separator.RIGHT_PAREN);
		dataList.add(Separator.RIGHT_PAREN);
		
		tokens.addAll(columnList);
		tokens.add(Keyword.VALUES);
		tokens.addAll(dataList);
		tokens.add(Separator.SEMICOLON);
		
		return new SqlStatementImpl(tokens);
	}
	

	/**
	 * 属性の出力戦略列挙型。
	 * 
	 * @author daisuke
	 */
	protected enum AttributeEmitStrategy {
		
		/** カラムの出力戦略 */
		COLUMN(ColumnModel.class) {
			
			@Override
			public List<Token> emit(AttributeModel attributeModel, DataTypeResolver dataTypeResolver,
					TokenResolver tokenResolver) {
				ReferenceResolver referenceResolver = attributeModel.getJiemamy().getReferenceResolver();
				ColumnModel columnModel = (ColumnModel) attributeModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				tokens.add(Identifier.of(columnModel.getName()));
				tokens.addAll(dataTypeResolver.resolveDataType(columnModel.getDataType(), referenceResolver));
				
				if (StringUtils.isEmpty(columnModel.getDefaultValue()) == false) {
					BuiltinDataType builtinDataType = columnModel.getDataType().toBuiltinDataType(referenceResolver);
					tokens.add(Keyword.DEFAULT);
					tokens.add(Literal
						.of(columnModel.getDefaultValue(), builtinDataType.getCategory().getLiteralType()));
				}
				
				NotNullConstraint nnModel = columnModel.getNotNullConstraint();
				if (nnModel != null) {
					if (StringUtils.isEmpty(nnModel.getName()) == false) {
						tokens.add(Keyword.CONSTRAINT);
						tokens.add(Identifier.of(nnModel.getName()));
					}
					tokens.add(Keyword.NOT);
					tokens.add(Keyword.NULL);
				}
				
				UniqueKey uniqueKey = columnModel.getUniqueKey();
				if (uniqueKey != null) {
					if (StringUtils.isEmpty(uniqueKey.getName()) == false) {
						tokens.add(Keyword.CONSTRAINT);
						tokens.add(Identifier.of(uniqueKey.getName()));
					}
					tokens.add(Keyword.UNIQUE);
				}
				
				PrimaryKey primaryKey = columnModel.getPrimaryKey();
				if (primaryKey != null) {
					if (StringUtils.isEmpty(primaryKey.getName()) == false) {
						tokens.add(Keyword.CONSTRAINT);
						tokens.add(Identifier.of(primaryKey.getName()));
					}
					tokens.add(Keyword.PRIMARY);
					tokens.add(Keyword.KEY);
				}
				
				ColumnCheckConstraint checkConstraint = columnModel.getCheckConstraint();
				if (checkConstraint != null) {
					if (StringUtils.isEmpty(checkConstraint.getName()) == false) {
						tokens.add(Keyword.CONSTRAINT);
						tokens.add(Identifier.of(checkConstraint.getName()));
					}
					tokens.add(Keyword.CHECK);
					tokens.add(Separator.LEFT_PAREN);
					tokens.add(Literal.of(checkConstraint.getExpression(), LiteralType.FRAGMENT));
					tokens.add(Separator.RIGHT_PAREN);
				}
				
				return tokens;
			}
		},
		
		/** 主キーの出力戦略 */
		PK(PrimaryKey.class) {
			
			@Override
			public List<Token> emit(AttributeModel attributeModel, DataTypeResolver dataTypeResolver,
					TokenResolver tokenResolver) {
				PrimaryKey primaryKey = (PrimaryKey) attributeModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				addConstraintNameDefinition(primaryKey, tokens);
				tokens.add(Keyword.PRIMARY);
				tokens.add(Keyword.KEY);
				ReferenceResolver referenceResolver = attributeModel.getJiemamy().getReferenceResolver();
				addColumnList(tokens, primaryKey.getKeyColumns(), referenceResolver);
				return tokens;
			}
			
		},
		
		/** ユニークキーの出力戦略 */
		UK(UniqueKey.class) {
			
			@Override
			public List<Token> emit(AttributeModel attributeModel, DataTypeResolver dataTypeResolver,
					TokenResolver tokenResolver) {
				UniqueKey uniqueKey = (UniqueKey) attributeModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				addConstraintNameDefinition(uniqueKey, tokens);
				tokens.add(Keyword.UNIQUE);
				tokens.add(Keyword.KEY);
				ReferenceResolver referenceResolver = attributeModel.getJiemamy().getReferenceResolver();
				addColumnList(tokens, uniqueKey.getKeyColumns(), referenceResolver);
				return tokens;
			}
		},
		
		/** 外部キーの出力戦略 */
		FK(ForeignKey.class) {
			
			@Override
			public List<Token> emit(AttributeModel attributeModel, DataTypeResolver dataTypeResolver,
					TokenResolver tokenResolver) {
				ReferenceResolver referenceResolver = attributeModel.getJiemamy().getReferenceResolver();
				ForeignKey foreignKey = (ForeignKey) attributeModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				addConstraintNameDefinition(foreignKey, tokens);
				tokens.add(Keyword.FOREIGN);
				tokens.add(Keyword.KEY);
				addColumnList(tokens, foreignKey.getKeyColumns(), referenceResolver);
				tokens.add(Keyword.REFERENCES);
				EntityModel referenceEntity = foreignKey.findReferencedEntity();
				tokens.add(Identifier.of(referenceEntity.getName()));
				addColumnList(tokens, foreignKey.getReferenceColumns(), referenceResolver);
				
				if (foreignKey.getMatchType() != null) {
					tokens.addAll(tokenResolver.resolve(foreignKey.getMatchType()));
				}
				if (foreignKey.getOnDelete() != null) {
					tokens.add(Keyword.ON);
					tokens.add(Keyword.DELETE);
					tokens.addAll(tokenResolver.resolve(foreignKey.getOnDelete()));
				}
				if (foreignKey.getOnUpdate() != null) {
					tokens.add(Keyword.ON);
					tokens.add(Keyword.UPDATE);
					tokens.addAll(tokenResolver.resolve(foreignKey.getOnUpdate()));
				}
				if (foreignKey.getDeferrability() != null) {
					Deferrability deferrability = foreignKey.getDeferrability();
					if (deferrability.isDeferrable() == false) {
						tokens.add(Keyword.NOT);
					}
					tokens.add(Keyword.DEFERRABLE);
					if (deferrability.getInitiallyCheckTime() != null) {
						tokens.addAll(tokenResolver.resolve(deferrability.getInitiallyCheckTime()));
					}
				}
				return tokens;
			}
		},
		
		/** CHECK制約の出力戦略 */
		TABLE_CHECK(TableCheckConstraint.class) {
			
			@Override
			public List<Token> emit(AttributeModel attributeModel, DataTypeResolver dataTypeResolver,
					TokenResolver tokenResolver) {
				CheckConstraint checkConstraint = (CheckConstraint) attributeModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				addConstraintNameDefinition(checkConstraint, tokens);
				tokens.add(Keyword.CHECK);
				tokens.add(Separator.LEFT_PAREN);
				tokens.add(Literal.of(checkConstraint.getExpression(), LiteralType.FRAGMENT));
				tokens.add(Separator.RIGHT_PAREN);
				return tokens;
			}
		};
		
		/**
		 * 属性モデルから出力戦略を取得する。
		 * 
		 * @param attribute 出力対象の属性モデル
		 * @return 出力戦略
		 */
		public static AttributeEmitStrategy fromAttribute(AttributeModel attribute) {
			for (AttributeEmitStrategy s : values()) {
				if (s.clazz == attribute.getClass()) {
					return s;
				}
				for (Class<?> c : attribute.getClass().getInterfaces()) {
					if (s.clazz == c) {
						return s;
					}
				}
			}
			return null;
		}
		
		private static void addColumnList(List<Token> tokens, List<ColumnRef> columnRefs,
				ReferenceResolver referenceResolver) {
			tokens.add(Separator.LEFT_PAREN);
			for (ColumnRef columnRef : columnRefs) {
				ColumnModel columnModel = referenceResolver.resolve(columnRef);
				tokens.add(Identifier.of(columnModel.getName()));
				tokens.add(Separator.COMMA);
			}
			
			if (columnRefs.isEmpty() == false) {
				tokens.remove(tokens.size() - 1);
			}
			tokens.add(Separator.RIGHT_PAREN);
		}
		
		private static void addConstraintNameDefinition(ConstraintModel constraint, List<Token> tokens) {
			if (StringUtils.isEmpty(constraint.getName()) == false) {
				tokens.add(Keyword.CONSTRAINT);
				tokens.add(Identifier.of(constraint.getName()));
			}
		}
		

		private final Class<? extends AttributeModel> clazz;
		

		AttributeEmitStrategy(Class<? extends AttributeModel> clazz) {
			this.clazz = clazz;
		}
		
		/**
		 * 属性モデルからトークンシーケンスを出力する。
		 * 
		 * @param attributeModel 属性モデル
		 * @param dataTypeResolver データ型リゾルバ
		 * @param tokenResolver トークンリゾルバ
		 * @param <T> 出力する属性モデルの型
		 * @return トークンシーケンス
		 */
		public abstract <T extends AttributeModel>List<Token> emit(T attributeModel, DataTypeResolver dataTypeResolver,
				TokenResolver tokenResolver);
		
	}
	
	/**
	 * エンティティの出力戦略。
	 * 
	 * @author daisuke
	 */
	protected enum EntityEmitStrategy {
		
		/** テーブルの出力戦略 */
		TABLE(TableModel.class) {
			
			@Override
			public SqlStatement emit(EntityModel entityModel, DataTypeResolver resolver, TokenResolver tokenResolver) {
				TableModel tableModel = (TableModel) entityModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				tokens.add(Keyword.CREATE);
				tokens.add(Keyword.TABLE);
				tokens.add(Identifier.of(tableModel.getName()));
				tokens.add(Separator.LEFT_PAREN);
				for (AttributeModel attributeModel : tableModel.getAttributes()) {
					if (attributeModel.hasAdapter(Disablable.class)
							&& Boolean.TRUE.equals(attributeModel.getAdapter(Disablable.class).isDisabled())) {
						continue;
					}
					AttributeEmitStrategy strategy = AttributeEmitStrategy.fromAttribute(attributeModel);
					List<Token> attributes = strategy.emit(attributeModel, resolver, tokenResolver);
					tokens.addAll(attributes);
					tokens.add(Separator.COMMA);
				}
				
				if (tableModel.getAttributes().isEmpty() == false) {
					tokens.remove(tokens.size() - 1);
				}
				tokens.add(Separator.RIGHT_PAREN);
				tokens.add(Separator.SEMICOLON);
				return new SqlStatementImpl(tokens);
			}
		},
		
		/** ビューの出力戦略 */
		VIEW(ViewModel.class) {
			
			@Override
			public SqlStatement emit(EntityModel entityModel, DataTypeResolver resolver, TokenResolver tokenResolver) {
				ViewModel viewModel = (ViewModel) entityModel;
				List<Token> tokens = CollectionsUtil.newArrayList();
				tokens.add(Keyword.CREATE);
				tokens.add(Keyword.VIEW);
				tokens.add(Identifier.of(viewModel.getName()));
				tokens.add(Keyword.AS);
				tokens.add(Literal.of(viewModel.getDefinition(), LiteralType.FRAGMENT));
				tokens.add(Separator.SEMICOLON);
				return new SqlStatementImpl(tokens);
			}
		};
		
		/**
		 * エンティティモデルから出力戦略を取得する。
		 * 
		 * @param entityModel 出力対象のエンティティモデル
		 * @return 出力戦略
		 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
		 */
		public static EntityEmitStrategy fromEntity(EntityModel entityModel) {
			Validate.notNull(entityModel);
			for (EntityEmitStrategy s : values()) {
				if (s.clazz == entityModel.getClass()) {
					return s;
				}
				for (Class<?> c : entityModel.getClass().getInterfaces()) {
					if (s.clazz == c) {
						return s;
					}
				}
			}
			return null;
		}
		

		private final Class<? extends EntityModel> clazz;
		

		EntityEmitStrategy(Class<? extends EntityModel> clazz) {
			this.clazz = clazz;
		}
		
		/**
		 * エンティティモデルからToken列を出力する。
		 * 
		 * @param entityModel エンティティモデル
		 * @param dataTypeResolver データ型リゾルバ
		 * @param tokenResolver トークンリゾルバ
		 * @return トークンシーケンス
		 */
		public abstract SqlStatement emit(EntityModel entityModel, DataTypeResolver dataTypeResolver,
				TokenResolver tokenResolver);
		
	}
}
