/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/03/03
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.facade;

import org.jiemamy.model.ConnectionProfile;
import org.jiemamy.model.DiagramPresentations;
import org.jiemamy.model.NodeProfile;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.connection.ConnectionAdapter;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.geometory.JmColor;
import org.jiemamy.model.geometory.JmPoint;
import org.jiemamy.model.geometory.JmRectangle;
import org.jiemamy.model.node.NodeAdapter;

/**
 * JiemamyのView仕様を含めたモデル操作を単純化するためのファサードインターフェイス。
 * 
 * @since 0.2
 * @author daisuke
 */
public interface JiemamyViewFacade extends JiemamyFacade {
	
	/**
	 * コネクションにベンドポイントを追加する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter 操作対象のコネクション
	 * @param bendpointIndex source側からtarget側に向かって数えたベンドポイントのインデックス
	 * @param position ベンドポイントの座標
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter, int bendpointIndex, JmPoint position);
	
	/**
	 * コネクションにビジュアル情報を設定する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter コネクション
	 * @param connectionProfile ビジュアル情報
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addConnectionProfile(int diagramIndex, ConnectionAdapter connectionAdapter, ConnectionProfile connectionProfile);
	
	/**
	 * {@link NodeProfile}を追加（表示状態にする）し、 {@link EntityModel}のラッピングノードであった場合は、{@link RootModel}に
	 * 対しての追加処理も行う。
	 * 
	 * <p>この操作に伴って、あらためて{@link #addEntity(EntityModel)}を実行する必要はない。</p>
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter ノードに対応する{@link NodeAdapter}
	 * @param rectangle {@code draw2D}のRectangleから{@link JmRectangle}に変換したもの
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addNode(int diagramIndex, NodeAdapter nodeAdapter, JmRectangle rectangle);
	
	/**
	 * {@link RootModel}に登録された{@link DiagramPresentations}アダプタに{@link NodeProfile}を追加する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter ノードに対応する{@link NodeAdapter}
	 * @param nodeProfile 追加する{@link NodeProfile}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void addNodeProfile(int diagramIndex, NodeAdapter nodeAdapter, NodeProfile nodeProfile);
	
	/**
	 * ノードの位置サイズを変更する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter 操作対象のノード
	 * @param boundary 新しい位置サイズ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void changeNodeBoundary(int diagramIndex, NodeAdapter nodeAdapter, JmRectangle boundary);
	
	/**
	 * ノードの色を変更する
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter 操作対象のノード
	 * @param color 新しい色
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void changeNodeColor(int diagramIndex, NodeAdapter nodeAdapter, JmColor color);
	
	/**
	 * ふたつの{@link TableModel}を関連づける{@link ForeignKey}を作成する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter コネクションに対応する{@link ConnectionAdapter}
	 * @param sourceTable 接続元のテーブル
	 * @param targetTable 接続先のテーブル
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void createConnection(int diagramIndex, ConnectionAdapter connectionAdapter, TableModel sourceTable,
			TableModel targetTable);
	
	/**
	 * ベンドポイントを移動する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter 操作対象のコネクション
	 * @param bendpointIndex source側からtarget側に向かって数えたベンドポイントのインデックス
	 * @param position 新しい座標
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void moveBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter, int bendpointIndex, JmPoint position);
	
	/**
	 * ベンドポイントを削除する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter 削除対象のコネクション
	 * @param bendpointIndex source側からtarget側に向かって数えたベンドポイントのインデックス
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter, int bendpointIndex);
	
	/**
	 * {@link RootModel}に登録された{@link DiagramPresentations}アダプタから{@link ConnectionProfile}を削除する。
	 * 
	 * <ul>
	 *   <li>コネクションが外部キーだった場合は、テーブルから外部キーも削除する。</li>
	 * </ul>
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter ノードに対応する{@link ConnectionAdapter}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeConnection(int diagramIndex, ConnectionAdapter connectionAdapter);
	
	/**
	 * {@link RootModel}に登録された{@link DiagramPresentations}アダプタから{@link ConnectionProfile}を削除する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter ノードに対応する{@link ConnectionAdapter}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeConnectionProfile(int diagramIndex, ConnectionAdapter connectionAdapter);
	
	/**
	 * ノードを削除する。
	 * 
	 * <ul>
	 *   <li>ノードがエンティティだった場合は、{@link RootModel}からエンティティも削除する。</li>
	 *   <li>このノードから出るコネクション及び、このノードへ向かうコネクションも同時に削除する。</li>
	 * </ul>
	 * 
	 * <p>GEFCommandで使用するためのよりファサードに近いメソッド。</p>
	 * <p>テストしやすくするためにj-eclipseから引っ張りだしてきた。</p>
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter 削除対象のノード
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeNode(int diagramIndex, NodeAdapter nodeAdapter);
	
	/**
	 * {@link RootModel}に登録された{@link DiagramPresentations}アダプタから{@link NodeProfile}を削除する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter ノードに対応する{@link NodeAdapter}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void removeNodeProfile(int diagramIndex, NodeAdapter nodeAdapter);
	
	/**
	 * ベンドポイントをリセットする。
	 * 
	 * <p>基本的にベンドポイントは全て削除されるが、自己接続コネクションであった場合は、可読性の向上のため、
	 * 2つ程度のベンドポイントを自動作成してもよい。</p>
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param connectionAdapter コネクション
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void resetBendpoint(int diagramIndex, ConnectionAdapter connectionAdapter);
	
	/**
	 * ノードの色を設定する。
	 * 
	 * @param diagramIndex ダイアグラムエディタのインデックス（エディタ内のタブインデックス）
	 * @param nodeAdapter ノード
	 * @param color 色
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 * @since 0.2
	 */
	void setColor(int diagramIndex, NodeAdapter nodeAdapter, JmColor color);
	
}
