/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/15
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.serializer;

import static org.jiemamy.xml.CoreQName.TABLE;
import static org.jiemamy.xml.CoreQName.VIEW;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.math.NumberUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.jiemamy.Jiemamy;
import org.jiemamy.JiemamyFactory;
import org.jiemamy.Version;
import org.jiemamy.dialect.BuiltinDataTypeMold;
import org.jiemamy.dialect.Dialect;
import org.jiemamy.dialect.internal.BuiltinDataTypeMoldImpl;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.internal.model.attribute.ColumnRefImpl;
import org.jiemamy.internal.model.datatype.DomainRefImpl;
import org.jiemamy.internal.model.entity.TableRefImpl;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.ColumnCheckConstraint;
import org.jiemamy.model.attribute.constraint.Deferrability;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.model.attribute.constraint.NotNullConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.TableCheckConstraint;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.attribute.constraint.Deferrability.InitiallyCheckTime;
import org.jiemamy.model.attribute.constraint.ForeignKey.ReferentialAction;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.BuiltinDataType;
import org.jiemamy.model.datatype.DataType;
import org.jiemamy.model.datatype.DataTypeCategory;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.model.entity.ViewModel;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.model.index.IndexColumnModel.SortOrder;
import org.jiemamy.serializer.SerializationException;
import org.jiemamy.utils.Adaptable;
import org.jiemamy.utils.DomUtil;
import org.jiemamy.utils.IterableNodeList;
import org.jiemamy.utils.SmartXpath;
import org.jiemamy.utils.UUIDUtil;
import org.jiemamy.xml.CoreQName;
import org.jiemamy.xml.JiemamyQName;

/**
 * Jiemamy XMLモデルのDOMから{@link RootModel}を生成するクラス。
 * 
 * @author daisuke
 */
class JiemamyDomParser {
	
	private static final String EL_CORE_NAME = "core:name";
	
	private static final String EL_CORE_LOGICALNAME = "core:logicalName";
	
	private static Logger logger = LoggerFactory.getLogger(JiemamyDomParser.class);
	
	private final Jiemamy jiemamy;
	
	private final JiemamyFactory factory;
	
	private final Document document;
	
	private SmartXpath smartXpath;
	

	/**
	 * インスタンスを生成する。
	 * @param jiemamy コンテキスト
	 * @param document 読み込むDOMドキュメント
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public JiemamyDomParser(Jiemamy jiemamy, Document document) {
		Validate.notNull(jiemamy);
		Validate.notNull(document);
		this.jiemamy = jiemamy;
		factory = jiemamy.getFactory();
		this.document = document;
		
		smartXpath = new SmartXpath(jiemamy);
	}
	
	/**
	 * DOMを解析し、{@link RootModel}を生成する。
	 * @return 生成した{@link RootModel}
	 * @throws SerializationException シリアライズに失敗した場合
	 */
	public RootModel parse() throws SerializationException {
		RootModel rootModel = parseRoot();
		parseDomainSection(rootModel);
		parseEntitiesSection(rootModel);
		parseDataSetSection(rootModel);
		
		return rootModel;
	}
	
	private DataType getDataType(Element element) {
		try {
			if (element.getChildNodes().getLength() == 0 && StringUtils.isEmpty(element.getAttribute("ref")) == false) {
				return new DomainRefImpl(jiemamy, getUUID(element, CoreQName.REF));
			} else {
				String typeCategoryString =
						(String) smartXpath.xpath.evaluate("core:typeCategory/text()", element, XPathConstants.STRING);
				String typeName =
						(String) smartXpath.xpath.evaluate("core:typeName/text()", element, XPathConstants.STRING);
				
				RootModel rootModel = factory.getRootModel();
				Dialect dialect;
				try {
					dialect = rootModel.findDialect();
				} catch (ClassNotFoundException e) {
					logger.warn("Dialectのロスト", e);
					dialect = null;
				}
				BuiltinDataTypeMold mold;
				if (StringUtils.isEmpty(typeCategoryString) == false) {
					if (dialect == null) {
						mold = new BuiltinDataTypeMoldImpl(DataTypeCategory.valueOf(typeCategoryString), typeName);
					} else {
						mold = dialect.findDataTypeMold(DataTypeCategory.valueOf(typeCategoryString), typeName);
					}
				} else {
					DataTypeCategory category;
					try {
						category = DataTypeCategory.valueOf(typeName);
					} catch (IllegalArgumentException e) {
						category = DataTypeCategory.INTEGER;
					}
					if (dialect == null) {
						mold = new BuiltinDataTypeMoldImpl(category, typeName);
					} else {
						mold = dialect.findDataTypeMold(category, typeName);
					}
				}
				BuiltinDataType dataType = factory.newDataType(mold);
				dataType.setTypeName(typeName);
				
				readAdapter(element, dataType);
				
				return dataType;
			}
		} catch (XPathExpressionException e) {
			throw new UnexpectedConditionError("XPath expression error", e);
		}
	}
	
	/**
	 * 要素の属性に設定された値を{@link UUID}として取得する。
	 * 
	 * @param domNode ノード
	 * @param node 属性指定
	 * @return {@link UUID}
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	private UUID getUUID(Node domNode, JiemamyQName node) {
		Validate.notNull(node);
		
		String uuidStr = smartXpath.getAttr(domNode, "@" + node.getQNameString()).getValue();
		return UUIDUtil.valueOfOrRandom(uuidStr);
	}
	
	private void parseAttributes(Element tableElement, TableModel tableModel) {
		NodeList attributeNodeList = smartXpath.getNodes(tableElement, "core:attributes/*");
		Iterable<Node> attributeIterableNodeList = new IterableNodeList(attributeNodeList);
		for (Node attributeNode : attributeIterableNodeList) {
			if ((attributeNode instanceof Element) == false) {
				continue;
			}
			Element attributeElement = (Element) attributeNode;
			if (DomUtil.isEqualsQName(attributeElement, CoreQName.COLUMN)) {
				parseColumn(attributeElement, tableModel);
			} else if (DomUtil.isEqualsQName(attributeElement, CoreQName.PRIMARY_KEY)) {
				parsePrimaryKey(attributeElement, tableModel);
			} else if (DomUtil.isEqualsQName(attributeElement, CoreQName.UNIQUE_KEY)) {
				parseUniqueKey(attributeElement, tableModel);
			} else if (DomUtil.isEqualsQName(attributeElement, CoreQName.FOREIGN_KEY)) {
				parseForeignKey(attributeElement, tableModel);
			} else if (DomUtil.isEqualsQName(attributeElement, CoreQName.TABLE_CHECK_CONSTRAINT)) {
				parseTableCheckConstraint(attributeElement, tableModel);
			} else {
				logger.warn("unknown attribute:" + attributeElement.getTagName());
			}
		}
	}
	
	private void parseColumn(Element columnElement, TableModel tableModel) {
		ColumnModel columnModel = factory.newModel(ColumnModel.class, getUUID(columnElement, CoreQName.ID));
		columnModel.setName(smartXpath.getTextContent(columnElement, EL_CORE_NAME));
		columnModel.setLogicalName(smartXpath.getTextContent(columnElement, EL_CORE_LOGICALNAME));
		columnModel.setDataType(getDataType((Element) smartXpath.getNode(columnElement, "core:dataType")));
		columnModel.setDefaultValue(smartXpath.getTextContent(columnElement, "core:defaultValue"));
		
		Element nnElement = (Element) smartXpath.getNode(columnElement, "core:notNull");
		if (nnElement != null) {
			NotNullConstraint nnModel = factory.newModel(NotNullConstraint.class, getUUID(nnElement, CoreQName.ID));
			nnModel.setName(smartXpath.getTextContent(nnElement, EL_CORE_NAME));
			readAdapter(nnElement, nnModel);
			columnModel.setNotNullConstraint(nnModel);
		}
		
		Element ukElement = (Element) smartXpath.getNode(columnElement, "core:columnUniqueKey");
		if (ukElement != null) {
			UniqueKey uniqueKey = factory.newModel(UniqueKey.class, getUUID(ukElement, CoreQName.ID));
			uniqueKey.setName(smartXpath.getTextContent(ukElement, EL_CORE_NAME));
			uniqueKey.getKeyColumns().add(factory.newReference(columnModel));
			readAdapter(ukElement, uniqueKey);
			columnModel.setUniqueKey(uniqueKey);
		}
		
		Element pkElement = (Element) smartXpath.getNode(columnElement, "core:columnPrimaryKey");
		if (pkElement != null) {
			PrimaryKey primaryKey = factory.newModel(PrimaryKey.class, getUUID(pkElement, CoreQName.ID));
			primaryKey.setName(smartXpath.getTextContent(pkElement, EL_CORE_NAME));
			primaryKey.getKeyColumns().add(factory.newReference(columnModel));
			readAdapter(pkElement, primaryKey);
			columnModel.setPrimaryKey(primaryKey);
		}
		
		Element checkElement = (Element) smartXpath.getNode(columnElement, "core:columnCheck");
		if (checkElement != null) {
			ColumnCheckConstraint checkConstraint =
					factory.newModel(ColumnCheckConstraint.class, getUUID(checkElement, CoreQName.ID));
			checkConstraint.setName(smartXpath.getTextContent(checkElement, EL_CORE_NAME));
			checkConstraint.setExpression(smartXpath.getTextContent(checkElement, "core:expression"));
			readAdapter(checkElement, checkConstraint);
			columnModel.setCheckConstraint(checkConstraint);
		}
		
		columnModel.setDescription(smartXpath.getTextContent(columnElement, "core:description"));
		readAdapter(columnElement, columnModel);
		tableModel.getAttributes().add(columnModel);
	}
	
	private void parseDataSetSection(RootModel rootModel) {
		NodeList dataSetNodeList = smartXpath.getNodes(document, "/core:jiemamy/core:dataSets/core:dataSet");
		Iterable<Node> dataSetIterableNodeList = new IterableNodeList(dataSetNodeList);
		for (Node dataSetNode : dataSetIterableNodeList) {
			
			DataSetModel dataSetModel = factory.newModel(DataSetModel.class, getUUID(dataSetNode, CoreQName.ID));
			dataSetModel.setName(smartXpath.getTextContent(dataSetNode, EL_CORE_NAME));
			
			NodeList tableRefNodeList = smartXpath.getNodes(dataSetNode, "core:tableRef");
			Iterable<Node> tableRefIterableNodeList = new IterableNodeList(tableRefNodeList);
			for (Node tableRefNode : tableRefIterableNodeList) {
				
				TableRef tableRef = new TableRefImpl(jiemamy, getUUID(tableRefNode, CoreQName.REF));
				List<RecordModel> recordModels = dataSetModel.getRecords().get(tableRef);
				if (recordModels == null) {
					recordModels = new ArrayList<RecordModel>();
				}
				
				NodeList recordNodeList = smartXpath.getNodes(tableRefNode, "core:record");
				Iterable<Node> recordIterableNodeList = new IterableNodeList(recordNodeList);
				for (Node recordNode : recordIterableNodeList) {
					RecordModel recordModel = factory.newModel(RecordModel.class, getUUID(recordNode, CoreQName.ID));
					
					NodeList columnRefNodeList = smartXpath.getNodes(recordNode, "core:columnRef");
					Iterable<Node> columnRefIterableNodeList = new IterableNodeList(columnRefNodeList);
					for (Node columnRefNode : columnRefIterableNodeList) {
						ColumnRef columnRef = new ColumnRefImpl(jiemamy, getUUID(columnRefNode, CoreQName.REF));
						recordModel.getValues().put(columnRef, columnRefNode.getTextContent());
					}
					recordModels.add(recordModel);
				}
				
				dataSetModel.getRecords().put(tableRef, recordModels);
			}
			rootModel.getDataSets().add(dataSetModel);
		}
	}
	
	private void parseDeferrability(Element keyElement, KeyConstraint keyModel) {
		Node deferrabilityElement = smartXpath.getNode(keyElement, "core:deferrability");
		
		if (deferrabilityElement != null) {
			Deferrability deferrability =
					factory.newModel(Deferrability.class, getUUID(deferrabilityElement, CoreQName.ID));
			String deferrableString = smartXpath.getTextContent(deferrabilityElement, "core:deferrable");
			deferrability.setDeferrable(BooleanUtils.toBoolean(deferrableString));
			
			String initiallyCheckTimeString =
					smartXpath.getTextContent(deferrabilityElement, "core:initiallyCheckTime");
			if (StringUtils.isEmpty(initiallyCheckTimeString) == false) {
				deferrability.setInitiallyCheckTime(InitiallyCheckTime.valueOf(initiallyCheckTimeString));
			}
			
			readAdapter((Element) deferrabilityElement, deferrability);
			keyModel.setDeferrability(deferrability);
		}
	}
	
	private void parseDomainSection(RootModel rootModel) {
		NodeList domainsNodeList = smartXpath.getNodes(document, "/core:jiemamy/core:domains/core:domain");
		Iterable<Node> domainsIterableNodeList = new IterableNodeList(domainsNodeList);
		for (Node domainNode : domainsIterableNodeList) {
			DomainModel domainModel = factory.newModel(DomainModel.class, getUUID(domainNode, CoreQName.ID));
			domainModel.setName(smartXpath.getTextContent(domainNode, EL_CORE_NAME));
			domainModel.setLogicalName(smartXpath.getTextContent(domainNode, EL_CORE_LOGICALNAME));
			domainModel.setDataType((BuiltinDataType) getDataType((Element) smartXpath.getNode(domainNode,
					"core:dataType")));
			
			Node nnNode = smartXpath.getNode(domainNode, "core:notNull");
			if (nnNode != null) {
				Element nnElement = (Element) nnNode;
				NotNullConstraint nnModel = factory.newModel(NotNullConstraint.class, getUUID(nnNode, CoreQName.ID));
				nnModel.setName(smartXpath.getTextContent(nnElement, EL_CORE_NAME));
				readAdapter(nnElement, nnModel);
				domainModel.setNotNullConstraint(nnModel);
			}
			
			Node checkNode = smartXpath.getNode(domainNode, "core:columnCheck");
			if (checkNode != null) {
				Element checkElement = (Element) checkNode;
				ColumnCheckConstraint checkConstraint =
						factory.newModel(ColumnCheckConstraint.class, getUUID(checkNode, CoreQName.ID));
				checkConstraint.setExpression(smartXpath.getTextContent(checkElement, "core:expression"));
				readAdapter(checkElement, checkConstraint);
				checkConstraint.setName(smartXpath.getTextContent(checkElement, EL_CORE_NAME));
				domainModel.setCheckConstraint(checkConstraint);
			}
			
			domainModel.setDescription(smartXpath.getTextContent(domainNode, "core:description"));
			
			readAdapter((Element) domainNode, domainModel);
			rootModel.getDomains().add(domainModel);
		}
	}
	
	private void parseEntitiesSection(RootModel rootModel) {
		NodeList entitityNodeList = smartXpath.getNodes(document, "/core:jiemamy/core:entities/*");
		Iterable<Node> entitityIterableNodeList = new IterableNodeList(entitityNodeList);
		for (Node entityNode : entitityIterableNodeList) {
			if ((entityNode instanceof Element) == false) {
				logger.warn("unexpected node (ignored): " + entityNode.getNodeType());
				continue;
			}
			Element entitiesElement = (Element) entityNode;
			EntityModel entityModel;
			if (entitiesElement.getNodeName().equals(TABLE.getQName().getLocalPart())) {
				entityModel = parseTable(entitiesElement);
			} else if (entitiesElement.getNodeName().equals(VIEW.getQName().getLocalPart())) {
				entityModel = parseView(entitiesElement);
			} else {
				logger.warn("unknown entity (ignored): " + entitiesElement.getNodeName());
				continue;
			}
			
			readAdapter(entitiesElement, entityModel);
			
			rootModel.getEntities().add(entityModel);
		}
	}
	
	private void parseForeignKey(Element fkElement, TableModel tableModel) {
		ForeignKey foreignKey = factory.newModel(ForeignKey.class, getUUID(fkElement, CoreQName.ID));
		foreignKey.setName(smartXpath.getTextContent(fkElement, EL_CORE_NAME));
		foreignKey.setLogicalName(smartXpath.getTextContent(fkElement, EL_CORE_LOGICALNAME));
		NodeList columnRefNodeList = smartXpath.getNodes(fkElement, "core:columnRefs/core:columnRef");
		Iterable<Node> columnRefIterableNodeList = new IterableNodeList(columnRefNodeList);
		for (Node node : columnRefIterableNodeList) {
			ColumnRef columnRef = new ColumnRefImpl(jiemamy, getUUID(node, CoreQName.REF));
			foreignKey.getKeyColumns().add(columnRef);
		}
		NodeList referenceColumnRefNodeList = smartXpath.getNodes(fkElement, "core:referenceColumns/core:columnRef");
		Iterable<Node> referenceColumnRefIterableNodeList = new IterableNodeList(referenceColumnRefNodeList);
		for (Node node : referenceColumnRefIterableNodeList) {
			ColumnRef columnRef = new ColumnRefImpl(jiemamy, getUUID(node, CoreQName.REF));
			foreignKey.getReferenceColumns().add(columnRef);
		}
		
		parseDeferrability(fkElement, foreignKey);
		
		String onDeleteString = smartXpath.getTextContent(fkElement, "core:onDelete");
		if (StringUtils.isEmpty(onDeleteString) == false) {
			foreignKey.setOnDelete(ReferentialAction.valueOf(onDeleteString));
		}
		
		String onUpdateString = smartXpath.getTextContent(fkElement, "core:onUpdate");
		if (StringUtils.isEmpty(onUpdateString) == false) {
			foreignKey.setOnUpdate(ReferentialAction.valueOf(onUpdateString));
		}
		
		readAdapter(fkElement, foreignKey);
		tableModel.getAttributes().add(foreignKey);
	}
	
	private void parseIndexes(Element tableElement, TableModel tableModel) {
		NodeList indexNodeList = smartXpath.getNodes(tableElement, "core:indexes/*");
		Iterable<Node> indexIterableNodeList = new IterableNodeList(indexNodeList);
		for (Node indexNode : indexIterableNodeList) {
			if ((indexNode instanceof Element) == false) {
				continue;
			}
			Element indexElement = (Element) indexNode;
			IndexModel indexModel = factory.newModel(IndexModel.class, getUUID(indexElement, CoreQName.ID));
			indexModel.setName(smartXpath.getTextContent(indexElement, EL_CORE_NAME));
			indexModel.setUnique(BooleanUtils.toBoolean(smartXpath.getTextContent(indexElement, "core:unique")));
			
			NodeList indexColumnNodeList = smartXpath.getNodes(indexElement, "core:indexColumns/core:indexColumn");
			Iterable<Node> indexColumnIterableNodeList = new IterableNodeList(indexColumnNodeList);
			for (Node node : indexColumnIterableNodeList) {
				IndexColumnModel indexColumnModel =
						factory.newModel(IndexColumnModel.class, getUUID(node, CoreQName.ID));
				Node columnRefNode = smartXpath.getNode(node, "core:columnRef");
				ColumnRef columnRef = new ColumnRefImpl(jiemamy, getUUID(columnRefNode, CoreQName.REF));
				indexColumnModel.setColumnRef(columnRef);
				indexModel.getIndexColumns().add(indexColumnModel);
				
				String sortOrderString = smartXpath.getTextContent(node, "core:sortOrder");
				if (StringUtils.isEmpty(sortOrderString) == false) {
					indexColumnModel.setSortOrder(SortOrder.valueOf(sortOrderString));
				}
				
			}
			tableModel.getIndexes().add(indexModel);
		}
	}
	
	private void parsePrimaryKey(Element pkElement, TableModel tableModel) {
		PrimaryKey primaryKey = factory.newModel(PrimaryKey.class, getUUID(pkElement, CoreQName.ID));
		primaryKey.setName(smartXpath.getTextContent(pkElement, EL_CORE_NAME));
		primaryKey.setLogicalName(smartXpath.getTextContent(pkElement, EL_CORE_LOGICALNAME));
		NodeList columnRefNodeList = smartXpath.getNodes(pkElement, "core:columnRefs/core:columnRef");
		Iterable<Node> columnRefIterableNodeList = new IterableNodeList(columnRefNodeList);
		for (Node node : columnRefIterableNodeList) {
			ColumnRef columnRef = new ColumnRefImpl(jiemamy, getUUID(node, CoreQName.REF));
			primaryKey.getKeyColumns().add(columnRef);
		}
		
		parseDeferrability(pkElement, primaryKey);
		
		readAdapter(pkElement, primaryKey);
		tableModel.getAttributes().add(primaryKey);
	}
	
	private RootModel parseRoot() throws SerializationException {
		Element rootElement = document.getDocumentElement();
		
		Attr verAttribute = smartXpath.getAttr(document, "/core:jiemamy/@version");
		String version = verAttribute.getValue();
		if (Version.SPEC.equals(version) == false) {
			throw new SerializationException("Version mismatch: serializer=" + Version.SPEC + ", xml=" + version);
		}
		RootModel rootModel = factory.getRootModel(getUUID(rootElement, CoreQName.ID));
		rootModel.setDialectClassName(smartXpath.getTextContent(document, "/core:jiemamy/core:dialect"));
		rootModel.setSchemaName(smartXpath.getTextContent(document, "/core:jiemamy/core:schemaName"));
		rootModel.setBeginScript(smartXpath.getTextContent(document, "/core:jiemamy/core:beginScript"));
		rootModel.setEndScript(smartXpath.getTextContent(document, "/core:jiemamy/core:endScript"));
		rootModel.setDescription(smartXpath.getTextContent(document, "/core:jiemamy/core:description"));
		
		return rootModel;
	}
	
	private TableModel parseTable(Element tableElement) {
		TableModel tableModel = factory.newModel(TableModel.class, getUUID(tableElement, CoreQName.ID));
		tableModel.setName(smartXpath.getTextContent(tableElement, EL_CORE_NAME));
		tableModel.setLogicalName(smartXpath.getTextContent(tableElement, EL_CORE_LOGICALNAME));
		tableModel.setBeginScript(smartXpath.getTextContent(tableElement, "core:beginScript"));
		tableModel.setEndScript(smartXpath.getTextContent(tableElement, "core:endScript"));
		tableModel.setDescription(smartXpath.getTextContent(tableElement, "core:description"));
		
		parseAttributes(tableElement, tableModel);
		parseIndexes(tableElement, tableModel);
		
		readAdapter(tableElement, tableModel);
		return tableModel;
	}
	
	private void parseTableCheckConstraint(Element checkElement, TableModel tableModel) {
		TableCheckConstraint checkConstraint =
				factory.newModel(TableCheckConstraint.class, getUUID(checkElement, CoreQName.ID));
		checkConstraint.setName(smartXpath.getTextContent(checkElement, EL_CORE_NAME));
		checkConstraint.setLogicalName(smartXpath.getTextContent(checkElement, EL_CORE_LOGICALNAME));
		checkConstraint.setExpression(smartXpath.getTextContent(checkElement, "core:expression"));
		readAdapter(checkElement, checkConstraint);
		tableModel.getAttributes().add(checkConstraint);
	}
	
	private void parseUniqueKey(Element ukElement, TableModel tableModel) {
		UniqueKey uniqueKey = factory.newModel(UniqueKey.class, getUUID(ukElement, CoreQName.ID));
		uniqueKey.setName(smartXpath.getTextContent(ukElement, EL_CORE_NAME));
		uniqueKey.setLogicalName(smartXpath.getTextContent(ukElement, EL_CORE_LOGICALNAME));
		NodeList columnRefNodeList = smartXpath.getNodes(ukElement, "core:columnRefs/core:columnRef");
		Iterable<Node> columnRefIterableNodeList = new IterableNodeList(columnRefNodeList);
		for (Node node : columnRefIterableNodeList) {
			ColumnRef columnRef = new ColumnRefImpl(jiemamy, getUUID(node, CoreQName.REF));
			uniqueKey.getKeyColumns().add(columnRef);
		}
		
		parseDeferrability(ukElement, uniqueKey);
		
		readAdapter(ukElement, uniqueKey);
		tableModel.getAttributes().add(uniqueKey);
	}
	
	private ViewModel parseView(Element viewElement) {
		ViewModel viewModel = factory.newModel(ViewModel.class, getUUID(viewElement, CoreQName.ID));
		viewModel.setName(smartXpath.getTextContent(viewElement, EL_CORE_NAME));
		viewModel.setLogicalName(smartXpath.getTextContent(viewElement, EL_CORE_LOGICALNAME));
		viewModel.setBeginScript(smartXpath.getTextContent(viewElement, "core:beginScript"));
		viewModel.setEndScript(smartXpath.getTextContent(viewElement, "core:endScript"));
		viewModel.setDescription(smartXpath.getTextContent(viewElement, "core:description"));
		viewModel.setDefinition(smartXpath.getTextContent(viewElement, "core:definition"));
		readAdapter(viewElement, viewModel);
		return viewModel;
	}
	
	private void readAdapter(Element element, Adaptable jiemamyElement) {
		NodeList adapterNodeList = smartXpath.getNodes(element, "core:adapter");
		Iterable<Node> adapterIterator = new IterableNodeList(adapterNodeList);
		
		for (Node adapterNode : adapterIterator) {
			Attr attr = smartXpath.getAttr(adapterNode, "@" + CoreQName.CLASS.getQNameString());
			String className = attr.getValue();
			try {
				Class<?> clazz = Class.forName(className);
				Object adapter;
				if (jiemamyElement.hasAdapter(clazz)) {
					adapter = jiemamyElement.getAdapter(clazz);
				} else {
					try {
						adapter = clazz.newInstance();
						jiemamyElement.registerAdapter(adapter);
					} catch (InstantiationException e) {
						logger.warn("ModelAdapter must have default constructor.", e);
						continue;
					} catch (IllegalAccessException e) {
						logger.warn("ModelAdapter must be public.", e);
						continue;
					}
				}
				NodeList propertyNodes = adapterNode.getChildNodes();
				Iterable<Node> propertyNodesIterator = new IterableNodeList(propertyNodes);
				
				for (Node propertyNode : propertyNodesIterator) {
					if ((propertyNode instanceof Element) == false) {
						continue;
					}
					Element propertyElement = (Element) propertyNode;
					String fieldName = null;
					try {
						fieldName = propertyElement.getTagName();
						Field field = adapter.getClass().getDeclaredField(fieldName);
						field.setAccessible(true);
						String value = propertyElement.getTextContent();
						Class<?> type = field.getType();
						if (type == Integer.class || type == int.class) {
							field.set(adapter, NumberUtils.toInt(value));
						} else if (type == Long.class || type == long.class) {
							field.set(adapter, NumberUtils.toLong(value));
						} else if (type == Float.class || type == float.class) {
							field.set(adapter, NumberUtils.toFloat(value));
						} else if (type == Double.class || type == double.class) {
							field.set(adapter, NumberUtils.toDouble(value));
						} else if (type == Boolean.class || type == boolean.class) {
							field.set(adapter, BooleanUtils.toBoolean(value));
						} else if (type == String.class) {
							field.set(adapter, value);
						} else {
							logger.warn("unsupported model adapter field data type: " + field.getType().getName());
						}
					} catch (SecurityException e) {
						logger.warn("security error", e);
					} catch (NoSuchFieldException e) {
						logger.warn("no such field " + fieldName + " in " + className, e);
					} catch (IllegalArgumentException e) {
						throw new UnexpectedConditionError("代入する型はifでチェックしているハズ。", e);
					} catch (IllegalAccessException e) {
						throw new UnexpectedConditionError("setAccessible(true)したハズ。", e);
					}
				}
				
			} catch (ClassNotFoundException e) {
				logger.warn("class not found (ignored): " + className, e);
			}
		}
	}
	
}
