/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/14
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.internal.serializer;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.lang.Validate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import org.jiemamy.Jiemamy;
import org.jiemamy.model.RootModel;
import org.jiemamy.serializer.JiemamySerializer;
import org.jiemamy.serializer.SerializationException;
import org.jiemamy.utils.LogMarker;

/**
 * シリアライザのDOMによる実装。
 * 
 * @author daisuke
 */
public class JiemamyDomSerializer implements JiemamySerializer {
	
	static final DocumentBuilderFactory DOCUMENT_BUILDER_FACTORY = DocumentBuilderFactory.newInstance();
	
	private static Logger logger = LoggerFactory.getLogger(JiemamyDomSerializer.class);
	
	/** インデント整形を行うためのXSLT */
	private static final String XSLT =
			"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" //
					+ "<xsl:stylesheet version=\"1.0\""
					+ " xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\""
					+ " xmlns:xalan=\"http://xml.apache.org/xslt\">"
					+ "<xsl:output method=\"xml\" encoding=\"UTF-8\" indent=\"yes\" xalan:indent-amount=\"2\"/>"
					+ "<xsl:template match=\"/\">" //
					+ "<xsl:copy-of select=\".\"/>" //
					+ "</xsl:template>" //
					+ "</xsl:stylesheet>";
	
	static {
		DOCUMENT_BUILDER_FACTORY.setNamespaceAware(true);
	}
	
	private final Jiemamy jiemamy;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param jiemamy コンテキスト
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public JiemamyDomSerializer(Jiemamy jiemamy) {
		Validate.notNull(jiemamy);
		this.jiemamy = jiemamy;
	}
	
	public RootModel deserialize(InputStream in) throws SerializationException {
		Validate.notNull(in);
		logger.debug(LogMarker.LIFECYCLE, "deserialize");
		
		RootModel rootModel;
		try {
			Document document = toDom(in);
			rootModel = toRootModel(document);
		} catch (ParserConfigurationException e) {
			throw new SerializationException(e);
		} catch (SAXException e) {
			throw new SerializationException(e);
		} catch (IOException e) {
			throw new SerializationException(e);
		}
		return rootModel;
	}
	
	public void serialize(RootModel rootModel, OutputStream out) throws SerializationException {
		Validate.notNull(rootModel);
		Validate.notNull(out);
		logger.debug(LogMarker.LIFECYCLE, "serialize");
		
		try {
			Document document = toDom(rootModel);
			
			DOMSource source = new DOMSource(document);
			TransformerFactory transFactory = TransformerFactory.newInstance();
			Transformer transformer =
					transFactory.newTransformer(new StreamSource(new ByteArrayInputStream(XSLT.getBytes())));
			
			// 書き出し
			transformer.transform(source, new StreamResult(out));
		} catch (DOMException e) {
			throw new SerializationException(e);
		} catch (TransformerConfigurationException e) {
			throw new SerializationException(e);
		} catch (ParserConfigurationException e) {
			throw new SerializationException(e);
		} catch (TransformerException e) {
			throw new SerializationException(e);
		}
	}
	
	/**
	 * XMLの{@link InputStream}より、DOMを組み立てる。
	 * 
	 * @param in 入力
	 * @return DOMドキュメント
	 * @throws ParserConfigurationException DocumentBuilderFactoryの設定異常
	 * @throws SAXException XMLのパースに失敗した場合
	 * @throws IOException 入出力エラーが発生した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected Document toDom(InputStream in) throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilder documentBuilder = DOCUMENT_BUILDER_FACTORY.newDocumentBuilder();
		Document document = documentBuilder.parse(in);
		
//		// XML SchemaのDOMを作る
//		DocumentBuilderFactory factory0 = DocumentBuilderFactory.newInstance();
//		DocumentBuilder builder0 = factory0.newDocumentBuilder();
//		Document coreXsd = builder0.parse(JiemamySerializer.class.getResourceAsStream("/jiemamy-core.xsd"));
//		Document viewXsd = builder0.parse(JiemamySerializer.class.getResourceAsStream("/jiemamy-view.xsd"));
//		
//		// XML Schemaオブジェクトを作る
//		SchemaFactory schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
//		Schema schema = schemaFactory.newSchema(new Source[] {
//			new DOMSource(coreXsd),
//			new DOMSource(viewXsd)
//		});
//		
//		// 妥当性検証
//		Validator validator = schema.newValidator();
//		validator.validate(new DOMSource(document));
		
		return document;
	}
	
	/**
	 * {@link RootModel}より、DOMを組み立てる。
	 * 
	 * @param rootModel 読み込む{@link RootModel}
	 * @return DOMドキュメント
	 * @throws ParserConfigurationException DocumentBuilderFactoryの設定異常
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected Document toDom(RootModel rootModel) throws ParserConfigurationException {
		return new JiemamyDomBuilder(rootModel).build();
	}
	
	/**
	 * DOMより、{@link RootModel}を組み立てる。
	 * 
	 * @param document 読み込むDOMドキュメント
	 * @return {@link RootModel}
	 * @throws SerializationException シリアライズに失敗した場合
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	protected RootModel toRootModel(Document document) throws SerializationException {
		return new JiemamyDomParser(jiemamy, document).parse();
	}
	
}
