/*
 * Copyright 2009 Jiemamy Project and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.enhancer.helper;

import java.lang.reflect.Array;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import javassist.CannotCompileException;
import javassist.ClassClassPath;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.NotFoundException;
import javassist.bytecode.Descriptor;

import org.jiemamy.utils.enhancer.EnhanceException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Javassistを利用して{@link java.lang.Class}と{@link javassist.CtClass}を相互に
 * 変換するためのライブラリ。
 * @version $Date: 2009-01-21 20:37:40 +0900 (水, 21  1月 2009) $
 * @author Suguru ARAKAWA (Gluegent, Inc.)
 */
public class JavassistConverter {
    
    private static final Logger LOG =
        LoggerFactory.getLogger(JavassistConverter.class);

    private static final Map<CtClass, Class<?>> PRIMITIVES;
    static {{
        Map<CtClass, Class<?>> map =
            new TreeMap<CtClass, Class<?>>(CtClassComparator.INSTANCE);
        map.put(CtClass.intType, int.class);
        map.put(CtClass.longType, long.class);
        map.put(CtClass.floatType, float.class);
        map.put(CtClass.doubleType, double.class);
        map.put(CtClass.shortType, short.class);
        map.put(CtClass.byteType, byte.class);
        map.put(CtClass.charType, char.class);
        map.put(CtClass.byteType, byte.class);
        map.put(CtClass.booleanType, boolean.class);
        PRIMITIVES = Collections.unmodifiableMap(map);
    }}
    
    private final ClassPool pool;
    
    private final ClassLoader loader;
    
    /**
     * インスタンスを生成する。
     * @param targetClass 基点とするクラス
     * @throws NullPointerException 引数に{@code null}が指定された場合
     */
    public JavassistConverter(Class<?> targetClass) {
        super();
        if (targetClass == null) {
            throw new NullPointerException("targetClass"); //$NON-NLS-1$
        }
        this.pool = new ClassPool();
        this.pool.appendClassPath(new ClassClassPath(targetClass));
        this.loader = new EmptyClassLoader(targetClass.getClassLoader());
    }
    

    /**
     * 指定の{@link java.lang.Class}オブジェクトを対応する{@link CtClass}に変換する。
     * @param klass 変換対象の{@link java.lang.Class}オブジェクト
     * @return 対応する{@link CtClass}
     * @throws EnhanceException 変換に失敗した場合
     */
    public CtClass toCtClass(Class<?> klass) throws EnhanceException {
        assert klass != null;
        assert klass.getCanonicalName() != null;
        String name = klass.getCanonicalName();
        try {
            return pool.get(name);
        }
        catch (NotFoundException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot load {0}",
                    name),
                e);
        }
    }

    /**
     * 指定の{@link CtClass}オブジェクトを対応する{@link java.lang.Class}オブジェクトへと
     * 変換して返す。
     * <p>
     * この操作が成功した場合、対象の{@link CtClass}は凍結される。
     * </p>
     * @param klass 対象の{@link CtClass}オブジェクト
     * @return 対応する{@link java.lang.Class}オブジェクト
     * @throws EnhanceException 変換に失敗した場合
     */
    public Class<?> toClass(CtClass klass) throws EnhanceException {
        assert klass != null;
        LOG.trace("Loading java.lang.Class: {}", klass.getName());
        try {
            if (klass.isPrimitive()) {
                Class<?> prim = toPrimitiveClass(klass);
                assert prim != null;
                return prim;
            }
            else if (klass.isArray()) {
                CtClass current = klass;
                int dim = 0;
                while (current.isArray()) {
                    dim++;
                    current = current.getComponentType();
                }
                Class<?> root = toClass(current);
                // FIXME to smart
                Object array = Array.newInstance(root, new int[dim]);
                return array.getClass();
            }
            else {
                Class<?> loaded = findFromLoader(klass);
                if (loaded != null) {
                    klass.freeze();
                    return loaded;
                }
                else {
                    LOG.debug("Register to JVM: {}", klass.getName());
                    return klass.toClass(
                        loader,
                        loader.getClass().getProtectionDomain());
                }
            }
        }
        catch (CannotCompileException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot load class (compile error):{0}",
                    klass.getName()),
                e);
        }
        catch (NotFoundException e) {
            throw new EnhanceException(
                MessageFormat.format(
                    "Cannot convert class (not found):{0}",
                    klass.getName()),
                e);
        }
    }

    /**
     * {@link #toClass(CtClass)}を配列に対して一斉適用する。
     * <p>
     * 返される配列に含まれるそれぞれの要素は、引数に渡した{@code CtClass}の一覧と同じ順序で
     * 対応する{@link java.lang.Class}オブジェクトが格納される。
     * </p>
     * @param classes 変換対象の一覧
     * @return 対応する変換結果の一覧
     * @throws EnhanceException 変換に失敗した場合
     */
    public Class<?>[] toClasses(CtClass[] classes) throws EnhanceException {
        assert classes != null;
        Class<?>[] results = new Class<?>[classes.length];
        for (int i = 0; i < results.length; i++) {
            results[i] = toClass(classes[i]);
        }
        return results;
    }


    /**
     * プリミティブ型を表現する指定の{@link CtClass}オブジェクトを
     * 対応する{@link java.lang.Class}オブジェクトへと変換して返す。
     * @param klass 対象のプリミティブ型を表現する{@link CtClass}オブジェクト
     * @return 対応する{@link java.lang.Class}オブジェクト
     */
    private Class<?> toPrimitiveClass(CtClass klass) {
        assert klass != null;
        assert klass.isPrimitive();
        Class<?> javaLangClass = PRIMITIVES.get(klass);
        if (javaLangClass == null) {
            throw new AssertionError("primitive ? " + klass);
        }
        return javaLangClass;
    }

    /**
     * 指定の{@link CtClass}オブジェクトに対応する{@link java.lang.Class}オブジェクトが
     * すでにロードされている場合、その値を返す。
     * <p>
     * エンハンサのクラスローダによって{@link java.lang.Class}オブジェクトがロードされていない場合、
     * この呼び出しは{@code null}を返す。
     * </p>
     * <p>
     * 引数に渡される{@link CtClass}は、プリミティブ型であってはならない。
     * </p>
     * @param klass 対象の{@link CtClass}オブジェクト
     * @return
     *      対応するロード済みの{@link java.lang.Class}オブジェクト、
     *      ロードされていない場合は{@code null}
     */
    private Class<?> findFromLoader(CtClass klass) {
        assert klass.isPrimitive() == false;
        try {
            if (klass.isArray()) {
                String descriptor = Descriptor.of(klass);
                return loader.loadClass(descriptor);
            }
            else {
                return loader.loadClass(klass.getName());
            }
        }
        catch (ClassNotFoundException e) {
            return null;
        }
    }

    /**
     * メソッドに対するアスペクトの一覧を対応するアドバイステーブルに変換する。
     * <p>
     * 返される配列に含まれる各要素は、引数に渡したアスペクトと同じ順序で、
     * それらの実行形式が格納される。
     * </p>
     * @param aspectList 変換するアスペクトの一覧
     * @return 対応するアドバイステーブル
     * @throws EnhanceException 変換に失敗した場合
     */
    public AdviceTable toMethodAspects(
            AspectList<CtMethod> aspectList) throws EnhanceException {
        
        assert aspectList != null;
        LOG.trace("Creating advice table entries: {}", aspectList);
        
        List<AdviceApplier> results = new ArrayList<AdviceApplier>();
        for (Aspect<CtMethod> aspect : aspectList) {
            CtMethod original = aspect.getOriginal();
            CtMethod actual = aspect.getActual();
            CtClass[] params;
            try {
                params = original.getParameterTypes();
            }
            catch (NotFoundException e) {
                throw new EnhanceException(
                    MessageFormat.format(
                        "Cannot resolve constructor {0}",
                        original.getName()),
                    e);
            }
            AdviceApplier helper = AdviceApplier.method(
                aspect.getHandlers(),
                toClass(original.getDeclaringClass()),
                original.getName(),
                toClass(actual.getDeclaringClass()),
                actual.getName(),
                toClasses(params));
            results.add(helper);
        }
        return new AdviceTable(aspectList, results);
    }

    /**
     * インスタンス生成に対するアスペクトの一覧を対応するアドバイステーブルに変換する。
     * <p>
     * 返される配列に含まれる各要素は、引数に渡したアスペクトと同じ順序で、
     * それらの実行形式が格納される。
     * </p>
     * @param aspectList 変換するアスペクトの一覧
     * @return 対応するアドバイステーブル
     * @throws EnhanceException 変換に失敗した場合
     */
    public AdviceTable toConstructorAspects(
            AspectList<CtConstructor> aspectList) throws EnhanceException {
        
        assert aspectList != null;
        LOG.trace("Creating advice table entries: {}", aspectList);
        
        List<AdviceApplier> results = new ArrayList<AdviceApplier>();
        for (Aspect<CtConstructor> aspect : aspectList) {
            CtConstructor original = aspect.getOriginal();
            CtConstructor actual = aspect.getActual();
            CtClass[] params;
            try {
                params = original.getParameterTypes();
            }
            catch (NotFoundException e) {
                throw new EnhanceException(
                    MessageFormat.format(
                        "Cannot resolve constructor {0}",
                        original.getName()),
                    e);
            }
            AdviceApplier helper = AdviceApplier.constructor(
                aspect.getHandlers(),
                toClass(original.getDeclaringClass()),
                toClass(actual.getDeclaringClass()),
                toClasses(params));
            results.add(helper);
        }
        return new AdviceTable(aspectList, results);
    }
}
