/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/02/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.utils.model;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.Validate;

import org.jiemamy.JiemamyProperty;
import org.jiemamy.exception.UnexpectedConditionError;
import org.jiemamy.model.JiemamyElement;
import org.jiemamy.utils.ReflectionUtil;

/**
 * {@link JiemamyElement}のプロパティを扱うユーティリティクラス。
 * 
 * @see JiemamyProperty
 * @author daisuke
 */
public final class JiemamyPropertyUtil {
	
	/**
	 * モデルからWidgetへの値転送時の、{@code null}に対する処理を行う。
	 * 
	 * @param value モデル保持値
	 * @return 入力が{@code null}の場合は空文字列、そうでない場合はそのまま出力する
	 */
	public static String careNull(String value) {
		return StringUtils.defaultIfEmpty(value, "");
	}
	
	/**
	 * Widgetからモデルへの値転送時の、{@code null}に対する処理を行う。
	 * 
	 * @param value 入力値
	 * @param nullable {@code null}が許されるモデルフィールドであるかどうか
	 * @return 入力が空（{@code null}または空文字）であり、nullableである場合は{@code null}、そうでない場合はそのまま出力する
	 */
	public static String careNull(String value, boolean nullable) {
		if (nullable) {
			return StringUtils.isEmpty(value) ? null : value;
		}
		return value;
	}
	
	/**
	 * {@link JiemamyElement}のプロパティの値を取得する。
	 * 
	 * @param <T> {@link JiemamyElement}の型
	 * @param element {@link JiemamyElement}インスタンス
	 * @param property プロパティを示す値
	 * @return プロパティの値
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public static <T extends JiemamyElement>Object getProperty(T element, JiemamyProperty<T> property) {
		Validate.notNull(element);
		Validate.notNull(property);
		
		String getterName = ReflectionUtil.convertFieldNameToAccessorName(property.name(), ReflectionUtil.GET);
		Method getterMethod;
		try {
			getterMethod = element.getClass().getMethod(getterName);
		} catch (SecurityException e) {
			throw new UnexpectedConditionError("security error", e);
		} catch (NoSuchMethodException e) {
			getterName = ReflectionUtil.convertFieldNameToAccessorName(property.name(), ReflectionUtil.IS);
			try {
				getterMethod = element.getClass().getMethod(getterName);
			} catch (SecurityException e2) {
				throw new UnexpectedConditionError("security error", e);
			} catch (NoSuchMethodException e2) {
				throw new UnexpectedConditionError("アクセサの存在はテストで保証されているはず", e);
			}
		}
		
		try {
			return getterMethod.invoke(element);
		} catch (IllegalArgumentException e) {
			throw new UnexpectedConditionError("引数なしのメソッドを拾っているはず", e);
		} catch (IllegalAccessException e) {
			throw new UnexpectedConditionError("getterはpublicなはず", e);
		} catch (InvocationTargetException e) {
			throw new UnexpectedConditionError("getter内で例外が発生", e);
		}
	}
	
	private JiemamyPropertyUtil() {
		throw new UnexpectedConditionError("不到達ポイント");
	}
}
