/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/21
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;
import java.util.regex.Pattern;

import org.apache.commons.lang.Validate;

import org.jiemamy.dialect.internal.ReservedWordsChecker;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * 全モデルをトラバースするバリデータの骨格実装。
 * 
 * @author daisuke
 */
public abstract class AbstractIdentifierValidator extends AbstractValidator {
	
	Collection<Problem> result = CollectionsUtil.newArrayList();
	
	/** 識別子としての妥当性を判断する正規表現パターン */
	private Pattern identifierPattern;
	
	/** 予約語プロバイダ */
	private ReservedWordsChecker reservedWords;
	

	/**
	 * インスタンスを生成する。
	 * 
	 * @param patternString 識別子としての妥当性を判断する正規表現パターン文字列
	 * @param reservedWords 予約語プロバイダ
	 * @throws IllegalArgumentException 引数に{@code null}を与えた場合
	 */
	public AbstractIdentifierValidator(String patternString, ReservedWordsChecker reservedWords) {
		Validate.notNull(patternString);
		Validate.notNull(reservedWords);
		
		identifierPattern = Pattern.compile(patternString);
		this.reservedWords = reservedWords;
	}
	
	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		result = super.validate(rootModel);
		
		for (DomainModel domainModel : rootModel.getDomains()) {
			isValid(domainModel.getName());
		}
		
		for (EntityModel entityModel : rootModel.getEntities()) {
			isValid(entityModel.getName());
			if (entityModel instanceof TableModel) {
				TableModel tableModel = (TableModel) entityModel;
				for (AttributeModel attributeModel : tableModel.getAttributes()) {
					isValid(attributeModel.getName());
				}
				
				for (IndexModel indexModel : tableModel.getIndexes()) {
					isValid(indexModel.getName());
				}
			}
		}
		
		return result;
	}
	
	boolean isValid(String name) {
		boolean valid = true;
		if (name == null) {
			return valid;
		}
		if (identifierPattern.matcher(name).matches() == false) {
			result.add(new InvalidNameProblem(name, identifierPattern));
			valid = false;
		}
		if (reservedWords.isReserved(name)) {
			result.add(new ReservedWordProblem(name));
			valid = false;
		}
		return valid;
	}
	

	static class InvalidNameProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param name 識別子名
		 * @param pattern 識別子が満たすべき正規表現パターン
		 */
		public InvalidNameProblem(String name, Pattern pattern) {
			super("E0010");
			registerMessage(Locale.JAPAN, "識別子{0}は、正規表現{1}にマッチしません");
			setArguments(new Object[] {
				name,
				pattern.pattern()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
	static class ReservedWordProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param name 識別子名
		 */
		public ReservedWordProblem(String name) {
			super("E0020");
			registerMessage(Locale.JAPAN, "予約語{0}が識別子として使用されています");
			setArguments(new Object[] {
				name
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
}
