/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/21
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.AttributeModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.CheckConstraint;
import org.jiemamy.model.attribute.constraint.ConstraintModel;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.model.attribute.constraint.PrimaryKey;
import org.jiemamy.model.attribute.constraint.UniqueKey;
import org.jiemamy.model.dataset.DataSetModel;
import org.jiemamy.model.dataset.RecordModel;
import org.jiemamy.model.datatype.DomainModel;
import org.jiemamy.model.entity.EntityModel;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.entity.TableRef;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * 全モデルをトラバースするバリデータの骨格実装。
 * 
 * @param <T> 識別子として使用する型
 * @author daisuke
 */
public abstract class AbstractTraversalValidator<T> extends AbstractValidator {
	
	private static final String TABLE = "table";
	

	private static String index(int i) {
		StringBuilder sb = new StringBuilder();
		sb.append("[").append(i).append("]");
		return sb.toString();
	}
	

	/** 結果として出力する問題点の集合 */
	protected Collection<Problem> result;
	
	/** 識別子：位置情報文字列のマップ */
	protected Map<T, String> ids;
	

	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		result = super.validate(rootModel);
		ids = CollectionsUtil.newHashMap();
		check(rootModel, "rootModel");
		
		int i = 0;
		for (DomainModel domainModel : rootModel.getDomains()) {
			check(domainModel, "domain[" + i + "]");
			check(domainModel.getDataType(), "domain" + index(i) + "/dataType");
			i++;
		}
		
		i = 0;
		for (EntityModel entityModel : rootModel.getEntities()) {
			check(entityModel, "entity" + index(i));
			if (entityModel instanceof TableModel) {
				TableModel tableModel = (TableModel) entityModel;
				int j = 0;
				for (AttributeModel attributeModel : tableModel.getAttributes()) {
					check(attributeModel, TABLE + index(i) + "/attribute" + index(j));
					if (attributeModel instanceof ColumnModel) {
						ColumnModel columnModel = (ColumnModel) attributeModel;
						check(columnModel.getDataType(), TABLE + index(i) + "/column" + index(j) + "/dataType");
						
						check(columnModel.getNotNullConstraint(), TABLE + index(i) + "/column" + index(j) + "/notNull");
						
						UniqueKey uniqueKey = columnModel.getUniqueKey();
						check(uniqueKey, TABLE + index(i) + "/column" + index(j) + "/uniqueKey");
						if (uniqueKey != null) {
							int k = 0;
							for (ColumnRef columnRef : uniqueKey.getKeyColumns()) {
								check(columnRef, TABLE + index(i) + "/column" + index(j) + "/uniqueKey/columnRef"
										+ index(k));
								k++;
							}
						}
						
						PrimaryKey primaryKey = columnModel.getPrimaryKey();
						check(primaryKey, TABLE + index(i) + "/column" + index(j) + "/primaryKey");
						if (primaryKey != null) {
							int k = 0;
							for (ColumnRef columnRef : primaryKey.getKeyColumns()) {
								check(columnRef, TABLE + index(i) + "/column" + index(j) + "/primaryKey/columnRef"
										+ index(k));
								k++;
							}
						}
						
						check(columnModel.getCheckConstraint(), TABLE + index(i) + "/column" + index(j) + "/check");
					} else if (attributeModel instanceof ConstraintModel) {
						checkConstraint(i, j, (ConstraintModel) attributeModel);
					}
					j++;
				}
				
				j = 0;
				for (IndexModel indexModel : tableModel.getIndexes()) {
					check(indexModel, TABLE + index(i) + "/index" + index(j));
					int k = 0;
					for (IndexColumnModel indexColumnModel : indexModel.getIndexColumns()) {
						check(indexColumnModel, TABLE + index(i) + "/index" + index(j) + "/indexColumn" + index(k));
						k++;
					}
					j++;
				}
			}
			i++;
		}
		
		i = 0;
		for (DataSetModel dataSetModel : rootModel.getDataSets()) {
			check(dataSetModel, "dataSet" + index(i));
			int j = 0;
			for (Map.Entry<TableRef, List<RecordModel>> entry : dataSetModel.getRecords().entrySet()) {
				check(entry.getKey(), "dataSet" + index(i) + "/tableRef");
				for (RecordModel recordModel : entry.getValue()) {
					check(recordModel, "dataSet" + index(i) + "/record" + index(j));
					j++;
				}
			}
			i++;
		}
		
		return result;
	}
	
	abstract void check(Object element, String pos);
	
	private void checkConstraint(int i, int j, ConstraintModel constraintModel) {
		if (constraintModel instanceof KeyConstraint) {
			if (constraintModel instanceof PrimaryKey) {
				PrimaryKey primaryKey = (PrimaryKey) constraintModel;
				int k = 0;
				for (ColumnRef columnRef : primaryKey.getKeyColumns()) {
					check(columnRef, TABLE + index(i) + "/pk" + index(j) + "/columnRef" + index(k));
					k++;
				}
			} else if (constraintModel instanceof UniqueKey) {
				UniqueKey uniqueKey = (UniqueKey) constraintModel;
				int k = 0;
				for (ColumnRef columnRef : uniqueKey.getKeyColumns()) {
					check(columnRef, TABLE + index(i) + "/uk" + index(j) + "/columnRef" + index(k));
					k++;
				}
			} else if (constraintModel instanceof ForeignKey) {
				ForeignKey foreignKey = (ForeignKey) constraintModel;
				int k = 0;
				for (ColumnRef columnRef : foreignKey.getKeyColumns()) {
					check(columnRef, TABLE + index(i) + "/fk" + index(j) + "/columnRef" + index(k));
					k++;
				}
				
				k = 0;
				for (ColumnRef columnRef : foreignKey.getReferenceColumns()) {
					check(columnRef, TABLE + index(i) + "/fk" + index(j) + "/refColumnRef" + index(k));
					k++;
				}
			}
		} else if (constraintModel instanceof CheckConstraint) {
			CheckConstraint checkConstraintModel = (CheckConstraint) constraintModel;
			check(checkConstraintModel.getExpression(), TABLE + index(i) + "/check" + index(j) + "/expression");
		}
	}
	
}
