/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.constraint.CheckConstraint;
import org.jiemamy.model.attribute.constraint.TableCheckConstraint;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * テーブルやカラムが保持するチェック制約に対するバリデータ。
 * 
 * <ul>
 *   <li>チェック制約には、制約式が設定されていなければならない。</li>
 * </ul>
 * 
 * @author daisuke
 */
public class CheckConstraintValidator extends AbstractValidator {
	
	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		Collection<TableModel> tableModels = rootModel.findEntities(TableModel.class);
		for (TableModel tableModel : tableModels) {
			for (ColumnModel columnModel : tableModel.findColumns()) {
				CheckConstraint checkConstraint = columnModel.getCheckConstraint();
				if (checkConstraint != null && verify(checkConstraint) == false) {
					result.add(new EmptyExpressionProblem(tableModel, columnModel));
				}
			}
			int index = 0;
			for (CheckConstraint checkConstraint : tableModel.findAttributes(TableCheckConstraint.class)) {
				if (verify(checkConstraint) == false) {
					result.add(new EmptyExpressionProblem(tableModel, index));
				}
				index++;
			}
		}
		return result;
	}
	
	/**
	 * チェック制約のexpressionに問題がないかどうか調べる。
	 * 
	 * @param checkConstraint 検査対象のチェック制約
	 * @return 問題がない場合は{@code true}、そうでない場合は{@code false}
	 */
	private boolean verify(CheckConstraint checkConstraint) {
		// TODO いつかは構文解析
		return StringUtils.isEmpty(checkConstraint.getExpression()) == false;
	}
	

	static class EmptyExpressionProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel 制約式未設定チェック制約が含まれるテーブル
		 * @param columnModel 制約式未設定チェック制約が設定されているカラム
		 */
		public EmptyExpressionProblem(TableModel tableModel, ColumnModel columnModel) {
			super("E0030");
			registerMessage(Locale.JAPAN, "テーブル{0}のカラム{1}に設定されたチェック制約に制約式がありません");
			setArguments(new Object[] {
				tableModel.getName(),
				columnModel.getName()
			});
		}
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel 制約式未設定チェック制約が含まれるテーブル
		 * @param index チェック制約のインデックス
		 */
		public EmptyExpressionProblem(TableModel tableModel, int index) {
			super("E0031");
			registerMessage(Locale.JAPAN, "テーブル{0}の{1}番目のチェック制約に制約式がありません");
			setArguments(new Object[] {
				tableModel.getName(),
				index + 1
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
}
