/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.constraint.ForeignKey;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.utils.model.RootModelUtil;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * {@link ForeignKey}の構成を調べるバリデータ。
 * 
 * <ul>
 *   <li>{@link ForeignKey#getKeyColumns()}と{@link ForeignKey#getReferenceColumns()}の要素数は一致していなければならない。</li>
 *   <li>{@link ForeignKey#getReferenceColumns()}は、参照先テーブルが持ついずれかの {@link KeyConstraint} のキー構成カラムと一致していなければならない。</li>
 * </ul>
 * 
 * @author daisuke
 */
public class ForeignKeyValidator extends AbstractValidator {
	
	private static Logger logger = LoggerFactory.getLogger(ForeignKeyValidator.class);
	

	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		for (ForeignKey foreignKey : RootModelUtil.getAllForeignKeys(rootModel)) {
			if (foreignKey.getKeyColumns().size() != foreignKey.getReferenceColumns().size()) {
				logger.error("fk(" + foreignKey.getId() + ") size error");
				result.add(new ReferenceMappingProblem(foreignKey));
			}
			
			if (foreignKey.findReferencedKeyConstraint() == null) {
				logger.error("fk(" + foreignKey.getId() + ") reference error");
				result.add(new ReferenceKeyProblem(foreignKey));
			}
		}
		return result;
	}
	

	static class ReferenceKeyProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param foreignKey 不正な外部キー
		 */
		public ReferenceKeyProblem(ForeignKey foreignKey) {
			super("E0080");
			registerMessage(Locale.JAPAN, "外部キー{0}が参照するカラムはキーになっていません");
			setArguments(new Object[] {
				StringUtils.isEmpty(foreignKey.getName()) ? foreignKey.getId().toString() : foreignKey.getName(),
				foreignKey.getKeyColumns().size(),
				foreignKey.getReferenceColumns().size()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
	static class ReferenceMappingProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param foreignKey 不正な外部キー
		 */
		public ReferenceMappingProblem(ForeignKey foreignKey) {
			super("E0090");
			registerMessage(Locale.JAPAN, "外部キー{0}のkeyColumnsサイズ({1})とreferenceColumnsサイズ({2})が一致していません");
			setArguments(new Object[] {
				foreignKey.getName() != null ? foreignKey.getName() : foreignKey.getId(),
				foreignKey.getKeyColumns().size(),
				foreignKey.getReferenceColumns().size()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
}
