/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;
import java.util.UUID;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.ReferenceResolver;
import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.model.index.IndexColumnModel;
import org.jiemamy.model.index.IndexModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * {@link IndexModel}の構成を調べるバリデータ。
 * 
 * @author daisuke
 */
public class IndexValidator extends AbstractValidator {
	
	private static Logger logger = LoggerFactory.getLogger(IndexValidator.class);
	

	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		ReferenceResolver referenceResolver = rootModel.getJiemamy().getReferenceResolver();
		for (TableModel tableModel : rootModel.findEntities(TableModel.class)) {
			Collection<UUID> columnIds = CollectionsUtil.newArrayList();
			for (ColumnModel columnModel : tableModel.findColumns()) {
				columnIds.add(columnModel.getId());
			}
			
			for (IndexModel indexModel : tableModel.getIndexes()) {
				Collection<UUID> referenceColumnIds = CollectionsUtil.newArrayList();
				if (indexModel.getIndexColumns().size() < 1) {
					result.add(new NoIndexColumnProblem(tableModel, indexModel));
				}
				for (IndexColumnModel indexColumnModel : indexModel.getIndexColumns()) {
					ColumnRef columnRef = indexColumnModel.getColumnRef();
					if (referenceColumnIds.contains(columnRef.getReferenceId())) {
						ColumnModel columnModel = referenceResolver.resolve(columnRef);
						result.add(new DuplicatedIndexColumnsProblem(indexModel, columnModel));
					}
					referenceColumnIds.add(columnRef.getReferenceId());
					if (columnIds.contains(columnRef.getReferenceId()) == false) {
						logger.error("illegal index (" + indexColumnModel.getId() + ")");
						result.add(new IllegalIndexColumnRefProblem(tableModel, indexModel, columnRef));
						
					}
				}
			}
		}
		return result;
	}
	

	static class DuplicatedIndexColumnsProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param indexModel 同じカラムが複数設定されたインデックス
		 * @param columnModel 複数設定されたカラム
		 */
		protected DuplicatedIndexColumnsProblem(IndexModel indexModel, ColumnModel columnModel) {
			super("E0100");
			registerMessage(Locale.JAPAN, "インデックス{0}に、同じカラム{1}が複数設定されています");
			setArguments(new Object[] {
				StringUtils.isEmpty(indexModel.getName()) ? indexModel.getId().toString() : indexModel.getName(),
				columnModel.getName()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
		
	}
	
	static class IllegalIndexColumnRefProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param columnRef 参照の切れた参照オブジェクト
		 * @param indexModel 不正参照オブジェクトを保持するインデックス
		 * @param tableModel インデックスを保持するテーブル
		 */
		private IllegalIndexColumnRefProblem(TableModel tableModel, IndexModel indexModel, ColumnRef columnRef) {
			super("E0110");
			registerMessage(Locale.JAPAN, "テーブル{0}にインデックス{1}が参照するカラム{2}は存在しません");
			setArguments(new Object[] {
				tableModel.getName(),
				indexModel.getName() != null ? indexModel.getName() : indexModel.getId().toString(),
				columnRef.getReferenceId().toString()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
	}
	
	static class NoIndexColumnProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel インデックスカラムを1つも持たないインデックスを持つテーブル
		 * @param indexModel インデックスカラムを1つも持たないインデックス
		 */
		public NoIndexColumnProblem(TableModel tableModel, IndexModel indexModel) {
			super("E0120");
			registerMessage(Locale.JAPAN, "テーブル{0}のインデックス{2}を構成するカラムが設定されていません");
			setArguments(new Object[] {
				StringUtils.isEmpty(tableModel.getName()) ? tableModel.getId().toString() : tableModel.getName(),
				StringUtils.isEmpty(indexModel.getName()) ? indexModel.getId().toString() : indexModel.getName(),
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
	}
}
