/*
 * Copyright 2007-2009 Jiemamy Project and the Others.
 * Created on 2009/01/26
 *
 * This file is part of Jiemamy.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.jiemamy.validator.impl;

import java.util.Collection;
import java.util.Locale;
import java.util.UUID;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.jiemamy.model.RootModel;
import org.jiemamy.model.attribute.ColumnModel;
import org.jiemamy.model.attribute.ColumnRef;
import org.jiemamy.model.attribute.constraint.KeyConstraint;
import org.jiemamy.model.entity.TableModel;
import org.jiemamy.utils.CollectionsUtil;
import org.jiemamy.validator.AbstractProblem;
import org.jiemamy.validator.AbstractValidator;
import org.jiemamy.validator.Problem;

/**
 * {@link KeyConstraint#getKeyColumns()}の構成を調べるバリデータ。
 * 
 * <ul>
 *   <li>keyColumnsは、自テーブルのカラムへの参照で構成されていなければならない。</li>
 * </ul>
 * 
 * @author daisuke
 */
public class KeyConstraintValidator extends AbstractValidator {
	
	private static Logger logger = LoggerFactory.getLogger(KeyConstraintValidator.class);
	

	@Override
	public Collection<Problem> validate(RootModel rootModel) {
		Collection<Problem> result = super.validate(rootModel);
		for (TableModel tableModel : rootModel.findEntities(TableModel.class)) {
			Collection<UUID> columnIds = CollectionsUtil.newArrayList();
			for (ColumnModel columnModel : tableModel.findColumns()) {
				columnIds.add(columnModel.getId());
			}
			for (KeyConstraint keyConstraint : tableModel.findAttributes(KeyConstraint.class)) {
				if (keyConstraint.getKeyColumns().size() < 1) {
					result.add(new NoKeyColumnProblem(tableModel, keyConstraint));
				}
				for (ColumnRef columnRef : keyConstraint.getKeyColumns()) {
					if (columnIds.contains(columnRef.getReferenceId()) == false) {
						logger.error("key(" + keyConstraint.getId() + ") columns error");
						result.add(new IllegalKeyColumnRefProblem(columnRef, keyConstraint, tableModel));
					}
				}
			}
		}
		return result;
	}
	

	static class IllegalKeyColumnRefProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param columnRef 参照の切れた参照オブジェクト
		 * @param keyConstraint 参照オブジェクトを保持するキー
		 * @param tableModel キーを保持するテーブル
		 */
		public IllegalKeyColumnRefProblem(ColumnRef columnRef, KeyConstraint keyConstraint, TableModel tableModel) {
			super("E0130");
			registerMessage(Locale.JAPAN, "テーブル{0}({1})にキー{2}({3})が参照するカラム{4}は存在しません");
			setArguments(new Object[] {
				tableModel.getName(),
				tableModel.getId().toString(),
				keyConstraint.getName(),
				keyConstraint.getId().toString(),
				columnRef.getReferenceId().toString()
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
	}
	
	static class NoKeyColumnProblem extends AbstractProblem {
		
		/**
		 * インスタンスを生成する。
		 * 
		 * @param tableModel キーカラムを1つも持たないキー制約を持つテーブル
		 * @param keyConstraint キーカラムを1つも持たないキー制約
		 */
		public NoKeyColumnProblem(TableModel tableModel, KeyConstraint keyConstraint) {
			super("E0140");
			registerMessage(Locale.JAPAN, "テーブル{0}のキー{2}を構成するカラムが設定されていません");
			setArguments(new Object[] {
				StringUtils.isEmpty(tableModel.getName()) ? tableModel.getId().toString() : tableModel.getName(),
				StringUtils.isEmpty(keyConstraint.getName()) ? keyConstraint.getId().toString() : keyConstraint
					.getName(),
			});
		}
		
		public Severity getSeverity() {
			return Severity.ERROR;
		}
	}
}
